#!/usr/bin/env python
#
# A library that provides a Python interface to the Telegram Bot API
# Copyright (C) 2015-2021
# Leandro Toledo de Souza <devs@python-telegram-bot.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].

import os
from pathlib import Path

import pytest
from flaky import flaky

from telegram import PhotoSize, Animation, Voice, TelegramError, MessageEntity, Bot
from telegram.error import BadRequest
from telegram.utils.helpers import escape_markdown
from tests.conftest import check_shortcut_call, check_shortcut_signature, check_defaults_handling


@pytest.fixture(scope='function')
def animation_file():
    f = open('tests/data/game.gif', 'rb')
    yield f
    f.close()


@pytest.fixture(scope='class')
def animation(bot, chat_id):
    with open('tests/data/game.gif', 'rb') as f:
        return bot.send_animation(
            chat_id, animation=f, timeout=50, thumb=open('tests/data/thumb.jpg', 'rb')
        ).animation


class TestAnimation:
    animation_file_id = 'CgADAQADngIAAuyVeEez0xRovKi9VAI'
    animation_file_unique_id = 'adc3145fd2e84d95b64d68eaa22aa33e'
    width = 320
    height = 180
    duration = 1
    # animation_file_url = 'https://python-telegram-bot.org/static/testfiles/game.gif'
    # Shortened link, the above one is cached with the wrong duration.
    animation_file_url = 'http://bit.ly/2L18jua'
    file_name = 'game.gif.mp4'
    mime_type = 'video/mp4'
    file_size = 4127
    caption = "Test *animation*"

    def test_creation(self, animation):
        assert isinstance(animation, Animation)
        assert isinstance(animation.file_id, str)
        assert isinstance(animation.file_unique_id, str)
        assert animation.file_id != ''
        assert animation.file_unique_id != ''

    def test_expected_values(self, animation):
        assert animation.file_size == self.file_size
        assert animation.mime_type == self.mime_type
        assert animation.file_name == self.file_name
        assert isinstance(animation.thumb, PhotoSize)

    @flaky(3, 1)
    @pytest.mark.timeout(10)
    def test_send_all_args(self, bot, chat_id, animation_file, animation, thumb_file):
        message = bot.send_animation(
            chat_id,
            animation_file,
            duration=self.duration,
            width=self.width,
            height=self.height,
            caption=self.caption,
            parse_mode='Markdown',
            disable_notification=False,
            thumb=thumb_file,
        )

        assert isinstance(message.animation, Animation)
        assert isinstance(message.animation.file_id, str)
        assert isinstance(message.animation.file_unique_id, str)
        assert message.animation.file_id != ''
        assert message.animation.file_unique_id != ''
        assert message.animation.file_name == animation.file_name
        assert message.animation.mime_type == animation.mime_type
        assert message.animation.file_size == animation.file_size
        assert message.animation.thumb.width == self.width
        assert message.animation.thumb.height == self.height

    @flaky(3, 1)
    @pytest.mark.timeout(10)
    def test_send_animation_custom_filename(self, bot, chat_id, animation_file, monkeypatch):
        def make_assertion(url, data, **kwargs):
            return data['animation'].filename == 'custom_filename'

        monkeypatch.setattr(bot.request, 'post', make_assertion)

        assert bot.send_animation(chat_id, animation_file, filename='custom_filename')

    @flaky(3, 1)
    @pytest.mark.timeout(10)
    def test_get_and_download(self, bot, animation):
        new_file = bot.get_file(animation.file_id)

        assert new_file.file_size == self.file_size
        assert new_file.file_id == animation.file_id
        assert new_file.file_path.startswith('https://')

        new_file.download('game.gif')

        assert os.path.isfile('game.gif')

    @flaky(3, 1)
    @pytest.mark.timeout(10)
    def test_send_animation_url_file(self, bot, chat_id, animation):
        message = bot.send_animation(
            chat_id=chat_id, animation=self.animation_file_url, caption=self.caption
        )

        assert message.caption == self.caption

        assert isinstance(message.animation, Animation)
        assert isinstance(message.animation.file_id, str)
        assert isinstance(message.animation.file_unique_id, str)
        assert message.animation.file_id != ''
        assert message.animation.file_unique_id != ''

        assert message.animation.duration == animation.duration
        assert message.animation.file_name == animation.file_name
        assert message.animation.mime_type == animation.mime_type
        assert message.animation.file_size == animation.file_size

    @flaky(3, 1)
    @pytest.mark.timeout(10)
    def test_send_animation_caption_entities(self, bot, chat_id, animation):
        test_string = 'Italic Bold Code'
        entities = [
            MessageEntity(MessageEntity.ITALIC, 0, 6),
            MessageEntity(MessageEntity.ITALIC, 7, 4),
            MessageEntity(MessageEntity.ITALIC, 12, 4),
        ]
        message = bot.send_animation(
            chat_id, animation, caption=test_string, caption_entities=entities
        )

        assert message.caption == test_string
        assert message.caption_entities == entities

    @flaky(3, 1)
    @pytest.mark.timeout(10)
    @pytest.mark.parametrize('default_bot', [{'parse_mode': 'Markdown'}], indirect=True)
    def test_send_animation_default_parse_mode_1(self, default_bot, chat_id, animation_file):
        test_string = 'Italic Bold Code'
        test_markdown_string = '_Italic_ *Bold* `Code`'

        message = default_bot.send_animation(chat_id, animation_file, caption=test_markdown_string)
        assert message.caption_markdown == test_markdown_string
        assert message.caption == test_string

    @flaky(3, 1)
    @pytest.mark.timeout(10)
    @pytest.mark.parametrize('default_bot', [{'parse_mode': 'Markdown'}], indirect=True)
    def test_send_animation_default_parse_mode_2(self, default_bot, chat_id, animation_file):
        test_markdown_string = '_Italic_ *Bold* `Code`'

        message = default_bot.send_animation(
            chat_id, animation_file, caption=test_markdown_string, parse_mode=None
        )
        assert message.caption == test_markdown_string
        assert message.caption_markdown == escape_markdown(test_markdown_string)

    @flaky(3, 1)
    @pytest.mark.timeout(10)
    @pytest.mark.parametrize('default_bot', [{'parse_mode': 'Markdown'}], indirect=True)
    def test_send_animation_default_parse_mode_3(self, default_bot, chat_id, animation_file):
        test_markdown_string = '_Italic_ *Bold* `Code`'

        message = default_bot.send_animation(
            chat_id, animation_file, caption=test_markdown_string, parse_mode='HTML'
        )
        assert message.caption == test_markdown_string
        assert message.caption_markdown == escape_markdown(test_markdown_string)

    def test_send_animation_local_files(self, monkeypatch, bot, chat_id):
        # For just test that the correct paths are passed as we have no local bot API set up
        test_flag = False
        expected = (Path.cwd() / 'tests/data/telegram.jpg/').as_uri()
        file = 'tests/data/telegram.jpg'

        def make_assertion(_, data, *args, **kwargs):
            nonlocal test_flag
            test_flag = data.get('animation') == expected and data.get('thumb') == expected

        monkeypatch.setattr(bot, '_post', make_assertion)
        bot.send_animation(chat_id, file, thumb=file)
        assert test_flag

    @flaky(3, 1)
    @pytest.mark.timeout(10)
    @pytest.mark.parametrize(
        'default_bot,custom',
        [
            ({'allow_sending_without_reply': True}, None),
            ({'allow_sending_without_reply': False}, None),
            ({'allow_sending_without_reply': False}, True),
        ],
        indirect=['default_bot'],
    )
    def test_send_animation_default_allow_sending_without_reply(
        self, default_bot, chat_id, animation, custom
    ):
        reply_to_message = default_bot.send_message(chat_id, 'test')
        reply_to_message.delete()
        if custom is not None:
            message = default_bot.send_animation(
                chat_id,
                animation,
                allow_sending_without_reply=custom,
                reply_to_message_id=reply_to_message.message_id,
            )
            assert message.reply_to_message is None
        elif default_bot.defaults.allow_sending_without_reply:
            message = default_bot.send_animation(
                chat_id, animation, reply_to_message_id=reply_to_message.message_id
            )
            assert message.reply_to_message is None
        else:
            with pytest.raises(BadRequest, match='message not found'):
                default_bot.send_animation(
                    chat_id, animation, reply_to_message_id=reply_to_message.message_id
                )

    @flaky(3, 1)
    @pytest.mark.timeout(10)
    def test_resend(self, bot, chat_id, animation):
        message = bot.send_animation(chat_id, animation.file_id)

        assert message.animation == animation

    def test_send_with_animation(self, monkeypatch, bot, chat_id, animation):
        def test(url, data, **kwargs):
            return data['animation'] == animation.file_id

        monkeypatch.setattr(bot.request, 'post', test)
        message = bot.send_animation(animation=animation, chat_id=chat_id)
        assert message

    def test_de_json(self, bot, animation):
        json_dict = {
            'file_id': self.animation_file_id,
            'file_unique_id': self.animation_file_unique_id,
            'width': self.width,
            'height': self.height,
            'duration': self.duration,
            'thumb': animation.thumb.to_dict(),
            'file_name': self.file_name,
            'mime_type': self.mime_type,
            'file_size': self.file_size,
        }
        animation = Animation.de_json(json_dict, bot)
        assert animation.file_id == self.animation_file_id
        assert animation.file_unique_id == self.animation_file_unique_id
        assert animation.thumb == animation.thumb
        assert animation.file_name == self.file_name
        assert animation.mime_type == self.mime_type
        assert animation.file_size == self.file_size

    def test_to_dict(self, animation):
        animation_dict = animation.to_dict()

        assert isinstance(animation_dict, dict)
        assert animation_dict['file_id'] == animation.file_id
        assert animation_dict['file_unique_id'] == animation.file_unique_id
        assert animation_dict['width'] == animation.width
        assert animation_dict['height'] == animation.height
        assert animation_dict['duration'] == animation.duration
        assert animation_dict['thumb'] == animation.thumb.to_dict()
        assert animation_dict['file_name'] == animation.file_name
        assert animation_dict['mime_type'] == animation.mime_type
        assert animation_dict['file_size'] == animation.file_size

    @flaky(3, 1)
    @pytest.mark.timeout(10)
    def test_error_send_empty_file(self, bot, chat_id):
        animation_file = open(os.devnull, 'rb')

        with pytest.raises(TelegramError):
            bot.send_animation(chat_id=chat_id, animation=animation_file)

    @flaky(3, 1)
    @pytest.mark.timeout(10)
    def test_error_send_empty_file_id(self, bot, chat_id):
        with pytest.raises(TelegramError):
            bot.send_animation(chat_id=chat_id, animation='')

    def test_error_send_without_required_args(self, bot, chat_id):
        with pytest.raises(TypeError):
            bot.send_animation(chat_id=chat_id)

    def test_get_file_instance_method(self, monkeypatch, animation):
        def make_assertion(*_, **kwargs):
            return kwargs['file_id'] == animation.file_id

        assert check_shortcut_signature(Animation.get_file, Bot.get_file, ['file_id'], [])
        assert check_shortcut_call(animation.get_file, animation.bot, 'get_file')
        assert check_defaults_handling(animation.get_file, animation.bot)

        monkeypatch.setattr(animation.bot, 'get_file', make_assertion)
        assert animation.get_file()

    def test_equality(self):
        a = Animation(
            self.animation_file_id,
            self.animation_file_unique_id,
            self.height,
            self.width,
            self.duration,
        )
        b = Animation('', self.animation_file_unique_id, self.height, self.width, self.duration)
        d = Animation('', '', 0, 0, 0)
        e = Voice(self.animation_file_id, self.animation_file_unique_id, 0)

        assert a == b
        assert hash(a) == hash(b)
        assert a is not b

        assert a != d
        assert hash(a) != hash(d)

        assert a != e
        assert hash(a) != hash(e)



#!/usr/bin/env python
#
# A library that provides a Python interface to the Telegram Bot API
# Copyright (C) 2015-2021
# Leandro Toledo de Souza <devs@python-telegram-bot.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].
import logging
import os
import subprocess
import sys
from io import BytesIO

from telegram import InputFile


class TestInputFile:
    png = os.path.join('tests', 'data', 'game.png')

    def test_subprocess_pipe(self):
        if sys.platform == 'win32':
            cmd = ['type', self.png]
        else:
            cmd = ['cat', self.png]

        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=(sys.platform == 'win32'))
        in_file = InputFile(proc.stdout)

        assert in_file.input_file_content == open(self.png, 'rb').read()
        assert in_file.mimetype == 'image/png'
        assert in_file.filename == 'image.png'

        try:
            proc.kill()
        except ProcessLookupError:
            # This exception may be thrown if the process has finished before we had the chance
            # to kill it.
            pass

    def test_mimetypes(self, caplog):
        # Only test a few to make sure logic works okay
        assert InputFile(open('tests/data/telegram.jpg', 'rb')).mimetype == 'image/jpeg'
        assert InputFile(open('tests/data/telegram.webp', 'rb')).mimetype == 'image/webp'
        assert InputFile(open('tests/data/telegram.mp3', 'rb')).mimetype == 'audio/mpeg'

        # Test guess from file
        assert InputFile(BytesIO(b'blah'), filename='tg.jpg').mimetype == 'image/jpeg'
        assert InputFile(BytesIO(b'blah'), filename='tg.mp3').mimetype == 'audio/mpeg'

        # Test fallback
        assert (
            InputFile(BytesIO(b'blah'), filename='tg.notaproperext').mimetype
            == 'application/octet-stream'
        )
        assert InputFile(BytesIO(b'blah')).mimetype == 'application/octet-stream'

        # Test string file
        with caplog.at_level(logging.DEBUG):
            assert InputFile(open('tests/data/text_file.txt')).mimetype == 'text/plain'

            assert len(caplog.records) == 1
            assert caplog.records[0].getMessage().startswith('Could not parse file content')

    def test_filenames(self):
        assert InputFile(open('tests/data/telegram.jpg', 'rb')).filename == 'telegram.jpg'
        assert InputFile(open('tests/data/telegram.jpg', 'rb'), filename='blah').filename == 'blah'
        assert (
            InputFile(open('tests/data/telegram.jpg', 'rb'), filename='blah.jpg').filename
            == 'blah.jpg'
        )
        assert InputFile(open('tests/data/telegram', 'rb')).filename == 'telegram'
        assert InputFile(open('tests/data/telegram', 'rb'), filename='blah').filename == 'blah'
        assert (
            InputFile(open('tests/data/telegram', 'rb'), filename='blah.jpg').filename
            == 'blah.jpg'
        )

        class MockedFileobject:
            # A open(?, 'rb') without a .name
            def __init__(self, f):
                self.f = open(f, 'rb')

            def read(self):
                return self.f.read()

        assert InputFile(MockedFileobject('tests/data/telegram.jpg')).filename == 'image.jpeg'
        assert (
            InputFile(MockedFileobject('tests/data/telegram.jpg'), filename='blah').filename
            == 'blah'
        )
        assert (
            InputFile(MockedFileobject('tests/data/telegram.jpg'), filename='blah.jpg').filename
            == 'blah.jpg'
        )
        assert (
            InputFile(MockedFileobject('tests/data/telegram')).filename
            == 'application.octet-stream'
        )
        assert (
            InputFile(MockedFileobject('tests/data/telegram'), filename='blah').filename == 'blah'
        )
        assert (
            InputFile(MockedFileobject('tests/data/telegram'), filename='blah.jpg').filename
            == 'blah.jpg'
        )

    def test_send_bytes(self, bot, chat_id):
        # We test this here and not at the respective test modules because it's not worth
        # duplicating the test for the different methods
        with open('tests/data/text_file.txt', 'rb') as file:
            message = bot.send_document(chat_id, file.read())

        out = BytesIO()
        assert message.document.get_file().download(out=out)
        out.seek(0)
        assert out.read().decode('utf-8') == 'PTB Rocks!'



#!/usr/bin/env python
#
# A library that provides a Python interface to the Telegram Bot API
# Copyright (C) 2015-2021
# Leandro Toledo de Souza <devs@python-telegram-bot.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].
import pickle
from collections import defaultdict

import pytest

from telegram import TelegramError, TelegramDecryptionError
from telegram.error import (
    Unauthorized,
    InvalidToken,
    NetworkError,
    BadRequest,
    TimedOut,
    ChatMigrated,
    RetryAfter,
    Conflict,
)


class TestErrors:
    def test_telegram_error(self):
        with pytest.raises(TelegramError, match="^test message$"):
            raise TelegramError("test message")
        with pytest.raises(TelegramError, match="^Test message$"):
            raise TelegramError("Error: test message")
        with pytest.raises(TelegramError, match="^Test message$"):
            raise TelegramError("[Error]: test message")
        with pytest.raises(TelegramError, match="^Test message$"):
            raise TelegramError("Bad Request: test message")

    def test_unauthorized(self):
        with pytest.raises(Unauthorized, match="test message"):
            raise Unauthorized("test message")
        with pytest.raises(Unauthorized, match="^Test message$"):
            raise Unauthorized("Error: test message")
        with pytest.raises(Unauthorized, match="^Test message$"):
            raise Unauthorized("[Error]: test message")
        with pytest.raises(Unauthorized, match="^Test message$"):
            raise Unauthorized("Bad Request: test message")

    def test_invalid_token(self):
        with pytest.raises(InvalidToken, match="Invalid token"):
            raise InvalidToken

    def test_network_error(self):
        with pytest.raises(NetworkError, match="test message"):
            raise NetworkError("test message")
        with pytest.raises(NetworkError, match="^Test message$"):
            raise NetworkError("Error: test message")
        with pytest.raises(NetworkError, match="^Test message$"):
            raise NetworkError("[Error]: test message")
        with pytest.raises(NetworkError, match="^Test message$"):
            raise NetworkError("Bad Request: test message")

    def test_bad_request(self):
        with pytest.raises(BadRequest, match="test message"):
            raise BadRequest("test message")
        with pytest.raises(BadRequest, match="^Test message$"):
            raise BadRequest("Error: test message")
        with pytest.raises(BadRequest, match="^Test message$"):
            raise BadRequest("[Error]: test message")
        with pytest.raises(BadRequest, match="^Test message$"):
            raise BadRequest("Bad Request: test message")

    def test_timed_out(self):
        with pytest.raises(TimedOut, match="^Timed out$"):
            raise TimedOut

    def test_chat_migrated(self):
        with pytest.raises(ChatMigrated, match="Group migrated to supergroup. New chat id: 1234"):
            raise ChatMigrated(1234)
        try:
            raise ChatMigrated(1234)
        except ChatMigrated as e:
            assert e.new_chat_id == 1234

    def test_retry_after(self):
        with pytest.raises(RetryAfter, match="Flood control exceeded. Retry in 12.0 seconds"):
            raise RetryAfter(12)

    def test_conflict(self):
        with pytest.raises(Conflict, match='Something something.'):
            raise Conflict('Something something.')

    @pytest.mark.parametrize(
        "exception, attributes",
        [
            (TelegramError("test message"), ["message"]),
            (Unauthorized("test message"), ["message"]),
            (InvalidToken(), ["message"]),
            (NetworkError("test message"), ["message"]),
            (BadRequest("test message"), ["message"]),
            (TimedOut(), ["message"]),
            (ChatMigrated(1234), ["message", "new_chat_id"]),
            (RetryAfter(12), ["message", "retry_after"]),
            (Conflict("test message"), ["message"]),
            (TelegramDecryptionError("test message"), ["message"]),
        ],
    )
    def test_errors_pickling(self, exception, attributes):
        pickled = pickle.dumps(exception)
        unpickled = pickle.loads(pickled)
        assert type(unpickled) is type(exception)
        assert str(unpickled) == str(exception)

        for attribute in attributes:
            assert getattr(unpickled, attribute) == getattr(exception, attribute)

    def test_pickling_test_coverage(self):
        """
        This test is only here to make sure that new errors will override __reduce__ properly.
        Add the new error class to the below covered_subclasses dict, if it's covered in the above
        test_errors_pickling test.
        """

        def make_assertion(cls):
            assert {sc for sc in cls.__subclasses__()} == covered_subclasses[cls]
            for subcls in cls.__subclasses__():
                make_assertion(subcls)

        covered_subclasses = defaultdict(set)
        covered_subclasses.update(
            {
                TelegramError: {
                    Unauthorized,
                    InvalidToken,
                    NetworkError,
                    ChatMigrated,
                    RetryAfter,
                    Conflict,
                    TelegramDecryptionError,
                },
                NetworkError: {BadRequest, TimedOut},
            }
        )

        make_assertion(TelegramError)



#!/usr/bin/env python
# pylint: disable=W0613, C0116
# type: ignore[union-attr]
# This program is dedicated to the public domain under the CC0 license.

"""
First, a few callback functions are defined. Then, those functions are passed to
the Dispatcher and registered at their respective places.
Then, the bot is started and runs until we press Ctrl-C on the command line.

Usage:
Example of a bot-user conversation using ConversationHandler.
Send /start to initiate the conversation.
Press Ctrl-C on the command line or send a signal to the process to stop the
bot.
"""

import logging
from typing import Dict

from telegram import ReplyKeyboardMarkup, Update
from telegram.ext import (
    Updater,
    CommandHandler,
    MessageHandler,
    Filters,
    ConversationHandler,
    CallbackContext,
)

# Enable logging
logging.basicConfig(
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s', level=logging.INFO
)

logger = logging.getLogger(__name__)

CHOOSING, TYPING_REPLY, TYPING_CHOICE = range(3)

reply_keyboard = [
    ['Age', 'Favourite colour'],
    ['Number of siblings', 'Something else...'],
    ['Done'],
]
markup = ReplyKeyboardMarkup(reply_keyboard, one_time_keyboard=True)


def facts_to_str(user_data: Dict[str, str]) -> str:
    facts = list()

    for key, value in user_data.items():
        facts.append(f'{key} - {value}')

    return "\n".join(facts).join(['\n', '\n'])


def start(update: Update, context: CallbackContext) -> int:
    update.message.reply_text(
        "Hi! My name is Doctor Botter. I will hold a more complex conversation with you. "
        "Why don't you tell me something about yourself?",
        reply_markup=markup,
    )

    return CHOOSING


def regular_choice(update: Update, context: CallbackContext) -> int:
    text = update.message.text
    context.user_data['choice'] = text
    update.message.reply_text(f'Your {text.lower()}? Yes, I would love to hear about that!')

    return TYPING_REPLY


def custom_choice(update: Update, context: CallbackContext) -> int:
    update.message.reply_text(
        'Alright, please send me the category first, ' 'for example "Most impressive skill"'
    )

    return TYPING_CHOICE


def received_information(update: Update, context: CallbackContext) -> int:
    user_data = context.user_data
    text = update.message.text
    category = user_data['choice']
    user_data[category] = text
    del user_data['choice']

    update.message.reply_text(
        "Neat! Just so you know, this is what you already told me:"
        f"{facts_to_str(user_data)} You can tell me more, or change your opinion"
        " on something.",
        reply_markup=markup,
    )

    return CHOOSING


def done(update: Update, context: CallbackContext) -> int:
    user_data = context.user_data
    if 'choice' in user_data:
        del user_data['choice']

    update.message.reply_text(
        f"I learned these facts about you: {facts_to_str(user_data)} Until next time!"
    )

    user_data.clear()
    return ConversationHandler.END


def main() -> None:
    # Create the Updater and pass it your bot's token.
    updater = Updater("TOKEN")

    # Get the dispatcher to register handlers
    dispatcher = updater.dispatcher

    # Add conversation handler with the states CHOOSING, TYPING_CHOICE and TYPING_REPLY
    conv_handler = ConversationHandler(
        entry_points=[CommandHandler('start', start)],
        states={
            CHOOSING: [
                MessageHandler(
                    Filters.regex('^(Age|Favourite colour|Number of siblings)$'), regular_choice
                ),
                MessageHandler(Filters.regex('^Something else...$'), custom_choice),
            ],
            TYPING_CHOICE: [
                MessageHandler(
                    Filters.text & ~(Filters.command | Filters.regex('^Done$')), regular_choice
                )
            ],
            TYPING_REPLY: [
                MessageHandler(
                    Filters.text & ~(Filters.command | Filters.regex('^Done$')),
                    received_information,
                )
            ],
        },
        fallbacks=[MessageHandler(Filters.regex('^Done$'), done)],
    )

    dispatcher.add_handler(conv_handler)

    # Start the Bot
    updater.start_polling()

    # Run the bot until you press Ctrl-C or the process receives SIGINT,
    # SIGTERM or SIGABRT. This should be used most of the time, since
    # start_polling() is non-blocking and will stop the bot gracefully.
    updater.idle()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# A library that provides a Python interface to the Telegram Bot API
# Copyright (C) 2015-2021
# Leandro Toledo de Souza <devs@python-telegram-bot.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].
# pylint: disable=W0622
"""This module contains the classes that represent Telegram PassportElementError."""

from typing import Any

from telegram import TelegramObject


class PassportElementError(TelegramObject):
    """Baseclass for the PassportElementError* classes.

    Objects of this class are comparable in terms of equality. Two objects of this class are
    considered equal, if their :attr:`source` and  :attr:`type`  are equal.

    Args:
        source (:obj:`str`): Error source.
        type (:obj:`str`): The section of the user's Telegram Passport which has the error.
        **kwargs (:obj:`dict`): Arbitrary keyword arguments.

    Attributes:
        source (:obj:`str`): Error source.
        type (:obj:`str`): The section of the user's Telegram Passport which has the error.
        message (:obj:`str`): Error message.

    """

    def __init__(self, source: str, type: str, message: str, **_kwargs: Any):
        # Required
        self.source = str(source)
        self.type = str(type)
        self.message = str(message)

        self._id_attrs = (self.source, self.type)


class PassportElementErrorDataField(PassportElementError):
    """
    Represents an issue in one of the data fields that was provided by the user. The error is
    considered resolved when the field's value changes.

    Objects of this class are comparable in terms of equality. Two objects of this class are
    considered equal, if their :attr:`source`, :attr:`type`, :attr:`field_name`, :attr:`data_hash`
    and  :attr:`message`  are equal.

    Args:
        type (:obj:`str`): The section of the user's Telegram Passport which has the error, one of
            "personal_details", "passport", "driver_license", "identity_card", "internal_passport",
            "address".
        field_name (:obj:`str`): Name of the data field which has the error.
        data_hash (:obj:`str`): Base64-encoded data hash.
        message (:obj:`str`): Error message.
        **kwargs (:obj:`dict`): Arbitrary keyword arguments.

    Attributes:
        type (:obj:`str`): The section of the user's Telegram Passport which has the error, one of
            "personal_details", "passport", "driver_license", "identity_card", "internal_passport",
            "address".
        field_name (:obj:`str`): Name of the data field which has the error.
        data_hash (:obj:`str`): Base64-encoded data hash.
        message (:obj:`str`): Error message.

    """

    def __init__(self, type: str, field_name: str, data_hash: str, message: str, **_kwargs: Any):
        # Required
        super().__init__('data', type, message)
        self.field_name = field_name
        self.data_hash = data_hash

        self._id_attrs = (self.source, self.type, self.field_name, self.data_hash, self.message)


class PassportElementErrorFile(PassportElementError):
    """
    Represents an issue with a document scan. The error is considered resolved when the file with
    the document scan changes.

    Objects of this class are comparable in terms of equality. Two objects of this class are
    considered equal, if their :attr:`source`, :attr:`type`, :attr:`file_hash`, :attr:`data_hash`
    and  :attr:`message`  are equal.

    Args:
        type (:obj:`str`): The section of the user's Telegram Passport which has the issue, one of
            "utility_bill", "bank_statement", "rental_agreement", "passport_registration",
            "temporary_registration".
        file_hash (:obj:`str`): Base64-encoded file hash.
        message (:obj:`str`): Error message.
        **kwargs (:obj:`dict`): Arbitrary keyword arguments.

    Attributes:
        type (:obj:`str`): The section of the user's Telegram Passport which has the issue, one of
            "utility_bill", "bank_statement", "rental_agreement", "passport_registration",
            "temporary_registration".
        file_hash (:obj:`str`): Base64-encoded file hash.
        message (:obj:`str`): Error message.

    """

    def __init__(self, type: str, file_hash: str, message: str, **_kwargs: Any):
        # Required
        super().__init__('file', type, message)
        self.file_hash = file_hash

        self._id_attrs = (self.source, self.type, self.file_hash, self.message)


class PassportElementErrorFiles(PassportElementError):
    """
    Represents an issue with a list of scans. The error is considered resolved when the file with
    the document scan changes.

    Objects of this class are comparable in terms of equality. Two objects of this class are
    considered equal, if their :attr:`source`, :attr:`type`, :attr:`file_hashes`, :attr:`data_hash`
    and  :attr:`message`  are equal.

    Args:
        type (:obj:`str`): The section of the user's Telegram Passport which has the issue, one of
            "utility_bill", "bank_statement", "rental_agreement", "passport_registration",
            "temporary_registration".
        file_hashes (List[:obj:`str`]): List of base64-encoded file hashes.
        message (:obj:`str`): Error message.
        **kwargs (:obj:`dict`): Arbitrary keyword arguments.

    Attributes:
        type (:obj:`str`): The section of the user's Telegram Passport which has the issue, one of
            "utility_bill", "bank_statement", "rental_agreement", "passport_registration",
            "temporary_registration".
        file_hashes (List[:obj:`str`]): List of base64-encoded file hashes.
        message (:obj:`str`): Error message.

    """

    def __init__(self, type: str, file_hashes: str, message: str, **_kwargs: Any):
        # Required
        super().__init__('files', type, message)
        self.file_hashes = file_hashes

        self._id_attrs = (self.source, self.type, self.message) + tuple(file_hashes)


class PassportElementErrorFrontSide(PassportElementError):
    """
    Represents an issue with the front side of a document. The error is considered resolved when
    the file with the front side of the document changes.

    Objects of this class are comparable in terms of equality. Two objects of this class are
    considered equal, if their :attr:`source`, :attr:`type`, :attr:`file_hash`, :attr:`data_hash`
    and  :attr:`message`  are equal.

    Args:
        type (:obj:`str`): The section of the user's Telegram Passport which has the issue, one of
            "passport", "driver_license", "identity_card", "internal_passport".
        file_hash (:obj:`str`): Base64-encoded hash of the file with the front side of the
            document.
        message (:obj:`str`): Error message.
        **kwargs (:obj:`dict`): Arbitrary keyword arguments.

    Attributes:
        type (:obj:`str`): The section of the user's Telegram Passport which has the issue, one of
            "passport", "driver_license", "identity_card", "internal_passport".
        file_hash (:obj:`str`): Base64-encoded hash of the file with the front side of the
            document.
        message (:obj:`str`): Error message.

    """

    def __init__(self, type: str, file_hash: str, message: str, **_kwargs: Any):
        # Required
        super().__init__('front_side', type, message)
        self.file_hash = file_hash

        self._id_attrs = (self.source, self.type, self.file_hash, self.message)


class PassportElementErrorReverseSide(PassportElementError):
    """
    Represents an issue with the front side of a document. The error is considered resolved when
    the file with the reverse side of the document changes.

    Objects of this class are comparable in terms of equality. Two objects of this class are
    considered equal, if their :attr:`source`, :attr:`type`, :attr:`file_hash`, :attr:`data_hash`
    and  :attr:`message`  are equal.

    Args:
        type (:obj:`str`): The section of the user's Telegram Passport which has the issue, one of
            "driver_license", "identity_card".
        file_hash (:obj:`str`): Base64-encoded hash of the file with the reverse side of the
            document.
        message (:obj:`str`): Error message.
        **kwargs (:obj:`dict`): Arbitrary keyword arguments.

    Attributes:
        type (:obj:`str`): The section of the user's Telegram Passport which has the issue, one of
            "passport", "driver_license", "identity_card", "internal_passport".
        file_hash (:obj:`str`): Base64-encoded hash of the file with the reverse side of the
            document.
        message (:obj:`str`): Error message.

    """

    def __init__(self, type: str, file_hash: str, message: str, **_kwargs: Any):
        # Required
        super().__init__('reverse_side', type, message)
        self.file_hash = file_hash

        self._id_attrs = (self.source, self.type, self.file_hash, self.message)


class PassportElementErrorSelfie(PassportElementError):
    """
    Represents an issue with the selfie with a document. The error is considered resolved when
    the file with the selfie changes.

    Objects of this class are comparable in terms of equality. Two objects of this class are
    considered equal, if their :attr:`source`, :attr:`type`, :attr:`file_hash`, :attr:`data_hash`
    and  :attr:`message`  are equal.

    Args:
        type (:obj:`str`): The section of the user's Telegram Passport which has the issue, one of
            "passport", "driver_license", "identity_card", "internal_passport".
        file_hash (:obj:`str`): Base64-encoded hash of the file with the selfie.
        message (:obj:`str`): Error message.
        **kwargs (:obj:`dict`): Arbitrary keyword arguments.

    Attributes:
        type (:obj:`str`): The section of the user's Telegram Passport which has the issue, one of
            "passport", "driver_license", "identity_card", "internal_passport".
        file_hash (:obj:`str`): Base64-encoded hash of the file with the selfie.
        message (:obj:`str`): Error message.

    """

    def __init__(self, type: str, file_hash: str, message: str, **_kwargs: Any):
        # Required
        super().__init__('selfie', type, message)
        self.file_hash = file_hash

        self._id_attrs = (self.source, self.type, self.file_hash, self.message)


class PassportElementErrorTranslationFile(PassportElementError):
    """
    Represents an issue with one of the files that constitute the translation of a document.
    The error is considered resolved when the file changes.

    Objects of this class are comparable in terms of equality. Two objects of this class are
    considered equal, if their :attr:`source`, :attr:`type`, :attr:`file_hash`, :attr:`data_hash`
    and  :attr:`message`  are equal.

    Args:
        type (:obj:`str`): Type of element of the user's Telegram Passport which has the issue,
            one of "passport", "driver_license", "identity_card", "internal_passport",
            "utility_bill", "bank_statement", "rental_agreement", "passport_registration",
            "temporary_registration".
        file_hash (:obj:`str`): Base64-encoded hash of the file.
        message (:obj:`str`): Error message.
        **kwargs (:obj:`dict`): Arbitrary keyword arguments.

    Attributes:
        type (:obj:`str`): Type of element of the user's Telegram Passport which has the issue,
            one of "passport", "driver_license", "identity_card", "internal_passport",
            "utility_bill", "bank_statement", "rental_agreement", "passport_registration",
            "temporary_registration".
        file_hash (:obj:`str`): Base64-encoded hash of the file.
        message (:obj:`str`): Error message.

    """

    def __init__(self, type: str, file_hash: str, message: str, **_kwargs: Any):
        # Required
        super().__init__('translation_file', type, message)
        self.file_hash = file_hash

        self._id_attrs = (self.source, self.type, self.file_hash, self.message)


class PassportElementErrorTranslationFiles(PassportElementError):
    """
    Represents an issue with the translated version of a document. The error is considered
    resolved when a file with the document translation change.

    Objects of this class are comparable in terms of equality. Two objects of this class are
    considered equal, if their :attr:`source`, :attr:`type`, :attr:`file_hashes`, :attr:`data_hash`
    and  :attr:`message`  are equal.

    Args:
        type (:obj:`str`): Type of element of the user's Telegram Passport which has the issue,
            one of "passport", "driver_license", "identity_card", "internal_passport",
            "utility_bill", "bank_statement", "rental_agreement", "passport_registration",
            "temporary_registration"
        file_hashes (List[:obj:`str`]): List of base64-encoded file hashes.
        message (:obj:`str`): Error message.
        **kwargs (:obj:`dict`): Arbitrary keyword arguments.

    Attributes:
        type (:obj:`str`): Type of element of the user's Telegram Passport which has the issue,
            one of "passport", "driver_license", "identity_card", "internal_passport",
            "utility_bill", "bank_statement", "rental_agreement", "passport_registration",
            "temporary_registration"
        file_hashes (List[:obj:`str`]): List of base64-encoded file hashes.
        message (:obj:`str`): Error message.

    """

    def __init__(self, type: str, file_hashes: str, message: str, **_kwargs: Any):
        # Required
        super().__init__('translation_files', type, message)
        self.file_hashes = file_hashes

        self._id_attrs = (self.source, self.type, self.message) + tuple(file_hashes)


class PassportElementErrorUnspecified(PassportElementError):
    """
    Represents an issue in an unspecified place. The error is considered resolved when new
    data is added.

    Objects of this class are comparable in terms of equality. Two objects of this class are
    considered equal, if their :attr:`source`, :attr:`type`, :attr:`element_hash`,
    :attr:`data_hash` and  :attr:`message`  are equal.

    Args:
        type (:obj:`str`): Type of element of the user's Telegram Passport which has the issue.
        element_hash (:obj:`str`): Base64-encoded element hash.
        message (:obj:`str`): Error message.
        **kwargs (:obj:`dict`): Arbitrary keyword arguments.

    Attributes:
        type (:obj:`str`): Type of element of the user's Telegram Passport which has the issue.
        element_hash (:obj:`str`): Base64-encoded element hash.
        message (:obj:`str`): Error message.

    """

    def __init__(self, type: str, element_hash: str, message: str, **_kwargs: Any):
        # Required
        super().__init__('unspecified', type, message)
        self.element_hash = element_hash

        self._id_attrs = (self.source, self.type, self.element_hash, self.message)



#!/usr/bin/env python
#
# A library that provides a Python interface to the Telegram Bot API
# Copyright (C) 2015-2021
# Leandro Toledo de Souza <devs@python-telegram-bot.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].
"""This module contains the class Updater, which tries to make creating Telegram bots intuitive."""

import logging
import ssl
import warnings
from queue import Queue
from signal import SIGABRT, SIGINT, SIGTERM, signal
from threading import Event, Lock, Thread, current_thread
from time import sleep
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union, no_type_check

from telegram import Bot, TelegramError
from telegram.error import InvalidToken, RetryAfter, TimedOut, Unauthorized
from telegram.ext import Dispatcher, JobQueue
from telegram.utils.deprecate import TelegramDeprecationWarning
from telegram.utils.helpers import get_signal_name
from telegram.utils.request import Request
from telegram.ext.utils.webhookhandler import WebhookAppClass, WebhookServer

if TYPE_CHECKING:
    from telegram.ext import BasePersistence, Defaults


class Updater:
    """
    This class, which employs the :class:`telegram.ext.Dispatcher`, provides a frontend to
    :class:`telegram.Bot` to the programmer, so they can focus on coding the bot. Its purpose is to
    receive the updates from Telegram and to deliver them to said dispatcher. It also runs in a
    separate thread, so the user can interact with the bot, for example on the command line. The
    dispatcher supports handlers for different kinds of data: Updates from Telegram, basic text
    commands and even arbitrary types. The updater can be started as a polling service or, for
    production, use a webhook to receive updates. This is achieved using the WebhookServer and
    WebhookHandler classes.

    Note:
        * You must supply either a :attr:`bot` or a :attr:`token` argument.
        * If you supply a :attr:`bot`, you will need to pass :attr:`defaults` to *both* the bot and
          the :class:`telegram.ext.Updater`.

    Args:
        token (:obj:`str`, optional): The bot's token given by the @BotFather.
        base_url (:obj:`str`, optional): Base_url for the bot.
        base_file_url (:obj:`str`, optional): Base_file_url for the bot.
        workers (:obj:`int`, optional): Amount of threads in the thread pool for functions
            decorated with ``@run_async`` (ignored if `dispatcher` argument is used).
        bot (:class:`telegram.Bot`, optional): A pre-initialized bot instance (ignored if
            `dispatcher` argument is used). If a pre-initialized bot is used, it is the user's
            responsibility to create it using a `Request` instance with a large enough connection
            pool.
        dispatcher (:class:`telegram.ext.Dispatcher`, optional): A pre-initialized dispatcher
            instance. If a pre-initialized dispatcher is used, it is the user's responsibility to
            create it with proper arguments.
        private_key (:obj:`bytes`, optional): Private key for decryption of telegram passport data.
        private_key_password (:obj:`bytes`, optional): Password for above private key.
        user_sig_handler (:obj:`function`, optional): Takes ``signum, frame`` as positional
            arguments. This will be called when a signal is received, defaults are (SIGINT,
            SIGTERM, SIGABRT) settable with :attr:`idle`.
        request_kwargs (:obj:`dict`, optional): Keyword args to control the creation of a
            `telegram.utils.request.Request` object (ignored if `bot` or `dispatcher` argument is
            used). The request_kwargs are very useful for the advanced users who would like to
            control the default timeouts and/or control the proxy used for http communication.
        use_context (:obj:`bool`, optional): If set to :obj:`True` uses the context based callback
            API (ignored if `dispatcher` argument is used). Defaults to :obj:`True`.
            **New users**: set this to :obj:`True`.
        persistence (:class:`telegram.ext.BasePersistence`, optional): The persistence class to
            store data that should be persistent over restarts (ignored if `dispatcher` argument is
            used).
        defaults (:class:`telegram.ext.Defaults`, optional): An object containing default values to
            be used if not set explicitly in the bot methods.

    Raises:
        ValueError: If both :attr:`token` and :attr:`bot` are passed or none of them.


    Attributes:
        bot (:class:`telegram.Bot`): The bot used with this Updater.
        user_sig_handler (:obj:`function`): Optional. Function to be called when a signal is
            received.
        update_queue (:obj:`Queue`): Queue for the updates.
        job_queue (:class:`telegram.ext.JobQueue`): Jobqueue for the updater.
        dispatcher (:class:`telegram.ext.Dispatcher`): Dispatcher that handles the updates and
            dispatches them to the handlers.
        running (:obj:`bool`): Indicates if the updater is running.
        persistence (:class:`telegram.ext.BasePersistence`): Optional. The persistence class to
            store data that should be persistent over restarts.
        use_context (:obj:`bool`): Optional. :obj:`True` if using context based callbacks.

    """

    _request = None

    def __init__(
        self,
        token: str = None,
        base_url: str = None,
        workers: int = 4,
        bot: Bot = None,
        private_key: bytes = None,
        private_key_password: bytes = None,
        user_sig_handler: Callable = None,
        request_kwargs: Dict[str, Any] = None,
        persistence: 'BasePersistence' = None,
        defaults: 'Defaults' = None,
        use_context: bool = True,
        dispatcher: Dispatcher = None,
        base_file_url: str = None,
    ):

        if defaults and bot:
            warnings.warn(
                'Passing defaults to an Updater has no effect when a Bot is passed '
                'as well. Pass them to the Bot instead.',
                TelegramDeprecationWarning,
                stacklevel=2,
            )

        if dispatcher is None:
            if (token is None) and (bot is None):
                raise ValueError('`token` or `bot` must be passed')
            if (token is not None) and (bot is not None):
                raise ValueError('`token` and `bot` are mutually exclusive')
            if (private_key is not None) and (bot is not None):
                raise ValueError('`bot` and `private_key` are mutually exclusive')
        else:
            if bot is not None:
                raise ValueError('`dispatcher` and `bot` are mutually exclusive')
            if persistence is not None:
                raise ValueError('`dispatcher` and `persistence` are mutually exclusive')
            if workers is not None:
                raise ValueError('`dispatcher` and `workers` are mutually exclusive')
            if use_context != dispatcher.use_context:
                raise ValueError('`dispatcher` and `use_context` are mutually exclusive')

        self.logger = logging.getLogger(__name__)

        if dispatcher is None:
            con_pool_size = workers + 4

            if bot is not None:
                self.bot = bot
                if bot.request.con_pool_size < con_pool_size:
                    self.logger.warning(
                        'Connection pool of Request object is smaller than optimal value (%s)',
                        con_pool_size,
                    )
            else:
                # we need a connection pool the size of:
                # * for each of the workers
                # * 1 for Dispatcher
                # * 1 for polling Updater (even if webhook is used, we can spare a connection)
                # * 1 for JobQueue
                # * 1 for main thread
                if request_kwargs is None:
                    request_kwargs = {}
                if 'con_pool_size' not in request_kwargs:
                    request_kwargs['con_pool_size'] = con_pool_size
                self._request = Request(**request_kwargs)
                self.bot = Bot(
                    token,  # type: ignore[arg-type]
                    base_url,
                    base_file_url=base_file_url,
                    request=self._request,
                    private_key=private_key,
                    private_key_password=private_key_password,
                    defaults=defaults,
                )
            self.update_queue: Queue = Queue()
            self.job_queue = JobQueue()
            self.__exception_event = Event()
            self.persistence = persistence
            self.dispatcher = Dispatcher(
                self.bot,
                self.update_queue,
                job_queue=self.job_queue,
                workers=workers,
                exception_event=self.__exception_event,
                persistence=persistence,
                use_context=use_context,
            )
            self.job_queue.set_dispatcher(self.dispatcher)
        else:
            con_pool_size = dispatcher.workers + 4

            self.bot = dispatcher.bot
            if self.bot.request.con_pool_size < con_pool_size:
                self.logger.warning(
                    'Connection pool of Request object is smaller than optimal value (%s)',
                    con_pool_size,
                )
            self.update_queue = dispatcher.update_queue
            self.__exception_event = dispatcher.exception_event
            self.persistence = dispatcher.persistence
            self.job_queue = dispatcher.job_queue
            self.dispatcher = dispatcher

        self.user_sig_handler = user_sig_handler
        self.last_update_id = 0
        self.running = False
        self.is_idle = False
        self.httpd = None
        self.__lock = Lock()
        self.__threads: List[Thread] = []

    def _init_thread(self, target: Callable, name: str, *args: object, **kwargs: object) -> None:
        thr = Thread(
            target=self._thread_wrapper,
            name=f"Bot:{self.bot.id}:{name}",
            args=(target,) + args,
            kwargs=kwargs,
        )
        thr.start()
        self.__threads.append(thr)

    def _thread_wrapper(self, target: Callable, *args: object, **kwargs: object) -> None:
        thr_name = current_thread().name
        self.logger.debug('%s - started', thr_name)
        try:
            target(*args, **kwargs)
        except Exception:
            self.__exception_event.set()
            self.logger.exception('unhandled exception in %s', thr_name)
            raise
        self.logger.debug('%s - ended', thr_name)

    def start_polling(
        self,
        poll_interval: float = 0.0,
        timeout: float = 10,
        clean: bool = False,
        bootstrap_retries: int = -1,
        read_latency: float = 2.0,
        allowed_updates: List[str] = None,
    ) -> Optional[Queue]:
        """Starts polling updates from Telegram.

        Args:
            poll_interval (:obj:`float`, optional): Time to wait between polling updates from
                Telegram in seconds. Default is 0.0.
            timeout (:obj:`float`, optional): Passed to :attr:`telegram.Bot.get_updates`.
            clean (:obj:`bool`, optional): Whether to clean any pending updates on Telegram servers
                before actually starting to poll. Default is :obj:`False`.
            bootstrap_retries (:obj:`int`, optional): Whether the bootstrapping phase of the
                `Updater` will retry on failures on the Telegram server.

                * < 0 - retry indefinitely (default)
                *   0 - no retries
                * > 0 - retry up to X times

            allowed_updates (List[:obj:`str`], optional): Passed to
                :attr:`telegram.Bot.get_updates`.
            read_latency (:obj:`float` | :obj:`int`, optional): Grace time in seconds for receiving
                the reply from server. Will be added to the `timeout` value and used as the read
                timeout from server (Default: 2).

        Returns:
            :obj:`Queue`: The update queue that can be filled from the main thread.

        """
        with self.__lock:
            if not self.running:
                self.running = True

                # Create & start threads
                self.job_queue.start()
                dispatcher_ready = Event()
                polling_ready = Event()
                self._init_thread(self.dispatcher.start, "dispatcher", ready=dispatcher_ready)
                self._init_thread(
                    self._start_polling,
                    "updater",
                    poll_interval,
                    timeout,
                    read_latency,
                    bootstrap_retries,
                    clean,
                    allowed_updates,
                    ready=polling_ready,
                )

                self.logger.debug('Waiting for Dispatcher and polling to start')
                dispatcher_ready.wait()
                polling_ready.wait()

                # Return the update queue so the main thread can insert updates
                return self.update_queue
            return None

    def start_webhook(
        self,
        listen: str = '127.0.0.1',
        port: int = 80,
        url_path: str = '',
        cert: str = None,
        key: str = None,
        clean: bool = False,
        bootstrap_retries: int = 0,
        webhook_url: str = None,
        allowed_updates: List[str] = None,
        force_event_loop: bool = False,
    ) -> Optional[Queue]:
        """
        Starts a small http server to listen for updates via webhook. If cert
        and key are not provided, the webhook will be started directly on
        http://listen:port/url_path, so SSL can be handled by another
        application. Else, the webhook will be started on
        https://listen:port/url_path

        Note:
            Due to an incompatibility of the Tornado library PTB uses for the webhook with Python
            3.8+ on Windows machines, PTB will attempt to set the event loop to
            :attr:`asyncio.SelectorEventLoop` and raise an exception, if an incompatible event loop
            has already been specified. See this `thread`_ for more details. To suppress the
            exception, set :attr:`force_event_loop` to :obj:`True`.

            .. _thread: https://github.com/tornadoweb/tornado/issues/2608

        Args:
            listen (:obj:`str`, optional): IP-Address to listen on. Default ``127.0.0.1``.
            port (:obj:`int`, optional): Port the bot should be listening on. Default ``80``.
            url_path (:obj:`str`, optional): Path inside url.
            cert (:obj:`str`, optional): Path to the SSL certificate file.
            key (:obj:`str`, optional): Path to the SSL key file.
            clean (:obj:`bool`, optional): Whether to clean any pending updates on Telegram servers
                before actually starting the webhook. Default is :obj:`False`.
            bootstrap_retries (:obj:`int`, optional): Whether the bootstrapping phase of the
                `Updater` will retry on failures on the Telegram server.

                * < 0 - retry indefinitely (default)
                *   0 - no retries
                * > 0 - retry up to X times

            webhook_url (:obj:`str`, optional): Explicitly specify the webhook url. Useful behind
                NAT, reverse proxy, etc. Default is derived from `listen`, `port` & `url_path`.
            allowed_updates (List[:obj:`str`], optional): Passed to
                :attr:`telegram.Bot.set_webhook`.
            force_event_loop (:obj:`bool`, optional): Force using the current event loop. See above
                note for details. Defaults to :obj:`False`

        Returns:
            :obj:`Queue`: The update queue that can be filled from the main thread.

        """
        with self.__lock:
            if not self.running:
                self.running = True

                # Create & start threads
                webhook_ready = Event()
                dispatcher_ready = Event()
                self.job_queue.start()
                self._init_thread(self.dispatcher.start, "dispatcher", dispatcher_ready)
                self._init_thread(
                    self._start_webhook,
                    "updater",
                    listen,
                    port,
                    url_path,
                    cert,
                    key,
                    bootstrap_retries,
                    clean,
                    webhook_url,
                    allowed_updates,
                    ready=webhook_ready,
                    force_event_loop=force_event_loop,
                )

                self.logger.debug('Waiting for Dispatcher and Webhook to start')
                webhook_ready.wait()
                dispatcher_ready.wait()

                # Return the update queue so the main thread can insert updates
                return self.update_queue
            return None

    @no_type_check
    def _start_polling(
        self,
        poll_interval,
        timeout,
        read_latency,
        bootstrap_retries,
        clean,
        allowed_updates,
        ready=None,
    ):  # pragma: no cover
        # Thread target of thread 'updater'. Runs in background, pulls
        # updates from Telegram and inserts them in the update queue of the
        # Dispatcher.

        self.logger.debug('Updater thread started (polling)')

        self._bootstrap(bootstrap_retries, clean=clean, webhook_url='', allowed_updates=None)

        self.logger.debug('Bootstrap done')

        def polling_action_cb():
            updates = self.bot.get_updates(
                self.last_update_id,
                timeout=timeout,
                read_latency=read_latency,
                allowed_updates=allowed_updates,
            )

            if updates:
                if not self.running:
                    self.logger.debug('Updates ignored and will be pulled again on restart')
                else:
                    for update in updates:
                        self.update_queue.put(update)
                    self.last_update_id = updates[-1].update_id + 1

            return True

        def polling_onerr_cb(exc):
            # Put the error into the update queue and let the Dispatcher
            # broadcast it
            self.update_queue.put(exc)

        if ready is not None:
            ready.set()

        self._network_loop_retry(
            polling_action_cb, polling_onerr_cb, 'getting Updates', poll_interval
        )

    @no_type_check
    def _network_loop_retry(self, action_cb, onerr_cb, description, interval):
        """Perform a loop calling `action_cb`, retrying after network errors.

        Stop condition for loop: `self.running` evaluates :obj:`False` or return value of
        `action_cb` evaluates :obj:`False`.

        Args:
            action_cb (:obj:`callable`): Network oriented callback function to call.
            onerr_cb (:obj:`callable`): Callback to call when TelegramError is caught. Receives the
                exception object as a parameter.
            description (:obj:`str`): Description text to use for logs and exception raised.
            interval (:obj:`float` | :obj:`int`): Interval to sleep between each call to
                `action_cb`.

        """
        self.logger.debug('Start network loop retry %s', description)
        cur_interval = interval
        while self.running:
            try:
                if not action_cb():
                    break
            except RetryAfter as exc:
                self.logger.info('%s', exc)
                cur_interval = 0.5 + exc.retry_after
            except TimedOut as toe:
                self.logger.debug('Timed out %s: %s', description, toe)
                # If failure is due to timeout, we should retry asap.
                cur_interval = 0
            except InvalidToken as pex:
                self.logger.error('Invalid token; aborting')
                raise pex
            except TelegramError as telegram_exc:
                self.logger.error('Error while %s: %s', description, telegram_exc)
                onerr_cb(telegram_exc)
                cur_interval = self._increase_poll_interval(cur_interval)
            else:
                cur_interval = interval

            if cur_interval:
                sleep(cur_interval)

    @staticmethod
    def _increase_poll_interval(current_interval: float) -> float:
        # increase waiting times on subsequent errors up to 30secs
        if current_interval == 0:
            current_interval = 1
        elif current_interval < 30:
            current_interval += current_interval / 2
        elif current_interval > 30:
            current_interval = 30
        return current_interval

    @no_type_check
    def _start_webhook(
        self,
        listen,
        port,
        url_path,
        cert,
        key,
        bootstrap_retries,
        clean,
        webhook_url,
        allowed_updates,
        ready=None,
        force_event_loop=False,
    ):
        self.logger.debug('Updater thread started (webhook)')
        use_ssl = cert is not None and key is not None
        if not url_path.startswith('/'):
            url_path = f'/{url_path}'

        # Create Tornado app instance
        app = WebhookAppClass(url_path, self.bot, self.update_queue)

        # Form SSL Context
        # An SSLError is raised if the private key does not match with the certificate
        if use_ssl:
            try:
                ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
                ssl_ctx.load_cert_chain(cert, key)
            except ssl.SSLError as exc:
                raise TelegramError('Invalid SSL Certificate') from exc
        else:
            ssl_ctx = None

        # Create and start server
        self.httpd = WebhookServer(listen, port, app, ssl_ctx)

        if use_ssl:
            # DO NOT CHANGE: Only set webhook if SSL is handled by library
            if not webhook_url:
                webhook_url = self._gen_webhook_url(listen, port, url_path)

            self._bootstrap(
                max_retries=bootstrap_retries,
                clean=clean,
                webhook_url=webhook_url,
                cert=open(cert, 'rb'),
                allowed_updates=allowed_updates,
            )
        elif clean:
            self.logger.warning(
                "cleaning updates is not supported if "
                "SSL-termination happens elsewhere; skipping"
            )

        self.httpd.serve_forever(force_event_loop=force_event_loop, ready=ready)

    @staticmethod
    def _gen_webhook_url(listen: str, port: int, url_path: str) -> str:
        return f'https://{listen}:{port}{url_path}'

    @no_type_check
    def _bootstrap(
        self, max_retries, clean, webhook_url, allowed_updates, cert=None, bootstrap_interval=5
    ):
        retries = [0]

        def bootstrap_del_webhook():
            self.bot.delete_webhook()
            return False

        def bootstrap_clean_updates():
            self.logger.debug('Cleaning updates from Telegram server')
            updates = self.bot.get_updates()
            while updates:
                updates = self.bot.get_updates(updates[-1].update_id + 1)
            return False

        def bootstrap_set_webhook():
            self.bot.set_webhook(
                url=webhook_url, certificate=cert, allowed_updates=allowed_updates
            )
            return False

        def bootstrap_onerr_cb(exc):
            if not isinstance(exc, Unauthorized) and (max_retries < 0 or retries[0] < max_retries):
                retries[0] += 1
                self.logger.warning(
                    'Failed bootstrap phase; try=%s max_retries=%s', retries[0], max_retries
                )
            else:
                self.logger.error('Failed bootstrap phase after %s retries (%s)', retries[0], exc)
                raise exc

        # Cleaning pending messages is done by polling for them - so we need to delete webhook if
        # one is configured.
        # We also take this chance to delete pre-configured webhook if this is a polling Updater.
        # NOTE: We don't know ahead if a webhook is configured, so we just delete.
        if clean or not webhook_url:
            self._network_loop_retry(
                bootstrap_del_webhook,
                bootstrap_onerr_cb,
                'bootstrap del webhook',
                bootstrap_interval,
            )
            retries[0] = 0

        # Clean pending messages, if requested.
        if clean:
            self._network_loop_retry(
                bootstrap_clean_updates,
                bootstrap_onerr_cb,
                'bootstrap clean updates',
                bootstrap_interval,
            )
            retries[0] = 0
            sleep(1)

        # Restore/set webhook settings, if needed. Again, we don't know ahead if a webhook is set,
        # so we set it anyhow.
        if webhook_url:
            self._network_loop_retry(
                bootstrap_set_webhook,
                bootstrap_onerr_cb,
                'bootstrap set webhook',
                bootstrap_interval,
            )

    def stop(self) -> None:
        """Stops the polling/webhook thread, the dispatcher and the job queue."""

        self.job_queue.stop()
        with self.__lock:
            if self.running or self.dispatcher.has_running_threads:
                self.logger.debug('Stopping Updater and Dispatcher...')

                self.running = False

                self._stop_httpd()
                self._stop_dispatcher()
                self._join_threads()

                # Stop the Request instance only if it was created by the Updater
                if self._request:
                    self._request.stop()

    @no_type_check
    def _stop_httpd(self) -> None:
        if self.httpd:
            self.logger.debug(
                'Waiting for current webhook connection to be '
                'closed... Send a Telegram message to the bot to exit '
                'immediately.'
            )
            self.httpd.shutdown()
            self.httpd = None

    @no_type_check
    def _stop_dispatcher(self) -> None:
        self.logger.debug('Requesting Dispatcher to stop...')
        self.dispatcher.stop()

    @no_type_check
    def _join_threads(self) -> None:
        for thr in self.__threads:
            self.logger.debug('Waiting for %s thread to end', thr.name)
            thr.join()
            self.logger.debug('%s thread has ended', thr.name)
        self.__threads = []

    @no_type_check
    def signal_handler(self, signum, frame) -> None:
        self.is_idle = False
        if self.running:
            self.logger.info(
                'Received signal %s (%s), stopping...', signum, get_signal_name(signum)
            )
            if self.persistence:
                # Update user_data, chat_data and bot_data before flushing
                self.dispatcher.update_persistence()
                self.persistence.flush()
            self.stop()
            if self.user_sig_handler:
                self.user_sig_handler(signum, frame)
        else:
            self.logger.warning('Exiting immediately!')
            # pylint: disable=C0415,W0212
            import os

            os._exit(1)

    def idle(self, stop_signals: Union[List, Tuple] = (SIGINT, SIGTERM, SIGABRT)) -> None:
        """Blocks until one of the signals are received and stops the updater.

        Args:
            stop_signals (:obj:`list` | :obj:`tuple`): List containing signals from the signal
                module that should be subscribed to. Updater.stop() will be called on receiving one
                of those signals. Defaults to (``SIGINT``, ``SIGTERM``, ``SIGABRT``).

        """
        for sig in stop_signals:
            signal(sig, self.signal_handler)

        self.is_idle = True

        while self.is_idle:
            sleep(1)



#!/usr/bin/env python
#
# A library that provides a Python interface to the Telegram Bot API
# Copyright (C) 2015-2021
# Leandro Toledo de Souza <devs@python-telegram-bot.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].

import pytest

from telegram import LabeledPrice, Location


@pytest.fixture(scope='class')
def labeled_price():
    return LabeledPrice(TestLabeledPrice.label, TestLabeledPrice.amount)


class TestLabeledPrice:
    label = 'label'
    amount = 100

    def test_expected_values(self, labeled_price):
        assert labeled_price.label == self.label
        assert labeled_price.amount == self.amount

    def test_to_dict(self, labeled_price):
        labeled_price_dict = labeled_price.to_dict()

        assert isinstance(labeled_price_dict, dict)
        assert labeled_price_dict['label'] == labeled_price.label
        assert labeled_price_dict['amount'] == labeled_price.amount

    def test_equality(self):
        a = LabeledPrice('label', 100)
        b = LabeledPrice('label', 100)
        c = LabeledPrice('Label', 101)
        d = Location(123, 456)

        assert a == b
        assert hash(a) == hash(b)

        assert a != c
        assert hash(a) != hash(c)

        assert a != d
        assert hash(a) != hash(d)



#!/usr/bin/env python
# pylint: disable=W0613, C0116
# type: ignore[union-attr]
# This program is dedicated to the public domain under the CC0 license.

"""
Simple Bot to reply to Telegram messages.

First, a few handler functions are defined. Then, those functions are passed to
the Dispatcher and registered at their respective places.
Then, the bot is started and runs until we press Ctrl-C on the command line.

Usage:
Basic Echobot example, repeats messages.
Press Ctrl-C on the command line or send a signal to the process to stop the
bot.
"""

import logging

from telegram import Update
from telegram.ext import Updater, CommandHandler, MessageHandler, Filters, CallbackContext

# Enable logging
logging.basicConfig(
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s', level=logging.INFO
)

logger = logging.getLogger(__name__)


# Define a few command handlers. These usually take the two arguments update and
# context. Error handlers also receive the raised TelegramError object in error.
def start(update: Update, context: CallbackContext) -> None:
    """Send a message when the command /start is issued."""
    update.message.reply_text('Hi!')


def help_command(update: Update, context: CallbackContext) -> None:
    """Send a message when the command /help is issued."""
    update.message.reply_text('Help!')


def echo(update: Update, context: CallbackContext) -> None:
    """Echo the user message."""
    update.message.reply_text(update.message.text)


def main():
    """Start the bot."""
    # Create the Updater and pass it your bot's token.
    updater = Updater("TOKEN")

    # Get the dispatcher to register handlers
    dispatcher = updater.dispatcher

    # on different commands - answer in Telegram
    dispatcher.add_handler(CommandHandler("start", start))
    dispatcher.add_handler(CommandHandler("help", help_command))

    # on noncommand i.e message - echo the message on Telegram
    dispatcher.add_handler(MessageHandler(Filters.text & ~Filters.command, echo))

    # Start the Bot
    updater.start_polling()

    # Run the bot until you press Ctrl-C or the process receives SIGINT,
    # SIGTERM or SIGABRT. This should be used most of the time, since
    # start_polling() is non-blocking and will stop the bot gracefully.
    updater.idle()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# A library that provides a Python interface to the Telegram Bot API
# Copyright (C) 2021
# Leandro Toledo de Souza <devs@python-telegram-bot.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].

"""
This file tests the case that PTB was installed *without* the optional dependency `passport`.
Currently this only means that cryptography is not installed.

Because imports in pytest are intricate, we just run
    pytest -k test_no_passport.py

with the TEST_NO_PASSPORT environment variable set in addition to the regular test suite.
Because actually uninstalling the optional dependencies would lead to errors in the test suite we
just mock the import to raise the expected exception.

Note that a fixture that just does this for every test that needs it is a nice idea, but for some
reason makes test_updater.py hang indefinitely on GitHub Actions (at least when Hinrich tried that)
"""
import os
from importlib import reload
from unittest import mock

import pytest

from telegram import bot
from telegram.passport import credentials
from tests.conftest import env_var_2_bool

TEST_NO_PASSPORT = env_var_2_bool(os.getenv('TEST_NO_PASSPORT', False))

if TEST_NO_PASSPORT:
    orig_import = __import__

    def import_mock(module_name, *args, **kwargs):
        if module_name.startswith('cryptography'):
            raise ModuleNotFoundError('We are testing without cryptography here')
        return orig_import(module_name, *args, **kwargs)

    with mock.patch('builtins.__import__', side_effect=import_mock):
        reload(bot)
        reload(credentials)


class TestNoPassport:
    """
    The monkeypatches simulate cryptography not being installed even when TEST_NO_PASSPORT is
    False, though that doesn't test the actual imports
    """

    def test_bot_init(self, bot_info, monkeypatch):
        if not TEST_NO_PASSPORT:
            monkeypatch.setattr(bot, 'CRYPTO_INSTALLED', False)
        with pytest.raises(RuntimeError, match='passport'):
            bot.Bot(bot_info['token'], private_key=1, private_key_password=2)

    def test_credentials_decrypt(self, monkeypatch):
        if not TEST_NO_PASSPORT:
            monkeypatch.setattr(credentials, 'CRYPTO_INSTALLED', False)
        with pytest.raises(RuntimeError, match='passport'):
            credentials.decrypt(1, 1, 1)

    def test_encrypted_credentials_decrypted_secret(self, monkeypatch):
        if not TEST_NO_PASSPORT:
            monkeypatch.setattr(credentials, 'CRYPTO_INSTALLED', False)
        ec = credentials.EncryptedCredentials('data', 'hash', 'secret')
        with pytest.raises(RuntimeError, match='passport'):
            ec.decrypted_secret



#!/usr/bin/env python
#
# A library that provides a Python interface to the Telegram Bot API
# Copyright (C) 2015-2021
# Leandro Toledo de Souza <devs@python-telegram-bot.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].
# pylint: disable=C0112, C0103, W0221
"""This module contains the Filters for use with the MessageHandler class."""

import re
import warnings

from abc import ABC, abstractmethod
from threading import Lock
from typing import (
    Dict,
    FrozenSet,
    List,
    Match,
    Optional,
    Pattern,
    Set,
    Tuple,
    Union,
    cast,
    NoReturn,
)

from telegram import Chat, Message, MessageEntity, Update, User

__all__ = [
    'Filters',
    'BaseFilter',
    'MessageFilter',
    'UpdateFilter',
    'InvertedFilter',
    'MergedFilter',
    'XORFilter',
]

from telegram.utils.deprecate import TelegramDeprecationWarning
from telegram.utils.types import SLT


class BaseFilter(ABC):
    """Base class for all Filters.

    Filters subclassing from this class can combined using bitwise operators:

    And:

        >>> (Filters.text & Filters.entity(MENTION))

    Or:

        >>> (Filters.audio | Filters.video)

    Exclusive Or:

        >>> (Filters.regex('To Be') ^ Filters.regex('Not 2B'))

    Not:

        >>> ~ Filters.command

    Also works with more than two filters:

        >>> (Filters.text & (Filters.entity(URL) | Filters.entity(TEXT_LINK)))
        >>> Filters.text & (~ Filters.forwarded)

    Note:
        Filters use the same short circuiting logic as python's `and`, `or` and `not`.
        This means that for example:

            >>> Filters.regex(r'(a?x)') | Filters.regex(r'(b?x)')

        With ``message.text == x``, will only ever return the matches for the first filter,
        since the second one is never evaluated.


    If you want to create your own filters create a class inheriting from either
    :class:`MessageFilter` or :class:`UpdateFilter` and implement a :meth:`filter` method that
    returns a boolean: :obj:`True` if the message should be
    handled, :obj:`False` otherwise.
    Note that the filters work only as class instances, not
    actual class objects (so remember to
    initialize your filter classes).

    By default the filters name (what will get printed when converted to a string for display)
    will be the class name. If you want to overwrite this assign a better name to the :attr:`name`
    class variable.

    Attributes:
        name (:obj:`str`): Name for this filter. Defaults to the type of filter.
        data_filter (:obj:`bool`): Whether this filter is a data filter. A data filter should
            return a dict with lists. The dict will be merged with
            :class:`telegram.ext.CallbackContext`'s internal dict in most cases
            (depends on the handler).
    """

    _name = None
    data_filter = False

    @abstractmethod
    def __call__(self, update: Update) -> Optional[Union[bool, Dict]]:
        pass

    def __and__(self, other: 'BaseFilter') -> 'BaseFilter':
        return MergedFilter(self, and_filter=other)

    def __or__(self, other: 'BaseFilter') -> 'BaseFilter':
        return MergedFilter(self, or_filter=other)

    def __xor__(self, other: 'BaseFilter') -> 'BaseFilter':
        return XORFilter(self, other)

    def __invert__(self) -> 'BaseFilter':
        return InvertedFilter(self)

    @property
    def name(self) -> Optional[str]:
        return self._name

    @name.setter
    def name(self, name: Optional[str]) -> None:
        self._name = name

    def __repr__(self) -> str:
        # We do this here instead of in a __init__ so filter don't have to call __init__ or super()
        if self.name is None:
            self.name = self.__class__.__name__
        return self.name


class MessageFilter(BaseFilter, ABC):
    """Base class for all Message Filters. In contrast to :class:`UpdateFilter`, the object passed
    to :meth:`filter` is ``update.effective_message``.

    Please see :class:`telegram.ext.filters.BaseFilter` for details on how to create custom
    filters.

    Attributes:
        name (:obj:`str`): Name for this filter. Defaults to the type of filter.
        data_filter (:obj:`bool`): Whether this filter is a data filter. A data filter should
            return a dict with lists. The dict will be merged with
            :class:`telegram.ext.CallbackContext`'s internal dict in most cases
            (depends on the handler).

    """

    def __call__(self, update: Update) -> Optional[Union[bool, Dict]]:
        return self.filter(update.effective_message)

    @abstractmethod
    def filter(self, message: Message) -> Optional[Union[bool, Dict]]:
        """This method must be overwritten.

        Args:
            message (:class:`telegram.Message`): The message that is tested.

        Returns:
            :obj:`dict` or :obj:`bool`

        """


class UpdateFilter(BaseFilter, ABC):
    """Base class for all Update Filters. In contrast to :class:`MessageFilter`, the object
    passed to :meth:`filter` is ``update``, which allows to create filters like
    :attr:`Filters.update.edited_message`.

    Please see :class:`telegram.ext.filters.BaseFilter` for details on how to create custom
    filters.

    Attributes:
        name (:obj:`str`): Name for this filter. Defaults to the type of filter.
        data_filter (:obj:`bool`): Whether this filter is a data filter. A data filter should
            return a dict with lists. The dict will be merged with
            :class:`telegram.ext.CallbackContext`'s internal dict in most cases
            (depends on the handler).

    """

    def __call__(self, update: Update) -> Optional[Union[bool, Dict]]:
        return self.filter(update)

    @abstractmethod
    def filter(self, update: Update) -> Optional[Union[bool, Dict]]:
        """This method must be overwritten.

        Args:
            update (:class:`telegram.Update`): The update that is tested.

        Returns:
            :obj:`dict` or :obj:`bool`.

        """


class InvertedFilter(UpdateFilter):
    """Represents a filter that has been inverted.

    Args:
        f: The filter to invert.

    """

    def __init__(self, f: BaseFilter):
        self.f = f

    def filter(self, update: Update) -> bool:
        return not bool(self.f(update))

    @property
    def name(self) -> str:
        return f"<inverted {self.f}>"

    @name.setter
    def name(self, name: str) -> NoReturn:
        raise RuntimeError('Cannot set name for InvertedFilter')


class MergedFilter(UpdateFilter):
    """Represents a filter consisting of two other filters.

    Args:
        base_filter: Filter 1 of the merged filter.
        and_filter: Optional filter to "and" with base_filter. Mutually exclusive with or_filter.
        or_filter: Optional filter to "or" with base_filter. Mutually exclusive with and_filter.

    """

    def __init__(
        self, base_filter: BaseFilter, and_filter: BaseFilter = None, or_filter: BaseFilter = None
    ):
        self.base_filter = base_filter
        if self.base_filter.data_filter:
            self.data_filter = True
        self.and_filter = and_filter
        if (
            self.and_filter
            and not isinstance(self.and_filter, bool)
            and self.and_filter.data_filter
        ):
            self.data_filter = True
        self.or_filter = or_filter
        if self.or_filter and not isinstance(self.and_filter, bool) and self.or_filter.data_filter:
            self.data_filter = True

    @staticmethod
    def _merge(base_output: Union[bool, Dict], comp_output: Union[bool, Dict]) -> Dict:
        base = base_output if isinstance(base_output, dict) else {}
        comp = comp_output if isinstance(comp_output, dict) else {}
        for k in comp.keys():
            # Make sure comp values are lists
            comp_value = comp[k] if isinstance(comp[k], list) else []
            try:
                # If base is a list then merge
                if isinstance(base[k], list):
                    base[k] += comp_value
                else:
                    base[k] = [base[k]] + comp_value
            except KeyError:
                base[k] = comp_value
        return base

    def filter(self, update: Update) -> Union[bool, Dict]:  # pylint: disable=R0911
        base_output = self.base_filter(update)
        # We need to check if the filters are data filters and if so return the merged data.
        # If it's not a data filter or an or_filter but no matches return bool
        if self.and_filter:
            # And filter needs to short circuit if base is falsey
            if base_output:
                comp_output = self.and_filter(update)
                if comp_output:
                    if self.data_filter:
                        merged = self._merge(base_output, comp_output)
                        if merged:
                            return merged
                    return True
        elif self.or_filter:
            # Or filter needs to short circuit if base is truthey
            if base_output:
                if self.data_filter:
                    return base_output
                return True

            comp_output = self.or_filter(update)
            if comp_output:
                if self.data_filter:
                    return comp_output
                return True
        return False

    @property
    def name(self) -> str:
        return (
            f"<{self.base_filter} {'and' if self.and_filter else 'or'} "
            f"{self.and_filter or self.or_filter}>"
        )

    @name.setter
    def name(self, name: str) -> NoReturn:
        raise RuntimeError('Cannot set name for MergedFilter')


class XORFilter(UpdateFilter):
    """Convenience filter acting as wrapper for :class:`MergedFilter` representing the an XOR gate
    for two filters.

    Args:
        base_filter: Filter 1 of the merged filter.
        xor_filter: Filter 2 of the merged filter.

    """

    def __init__(self, base_filter: BaseFilter, xor_filter: BaseFilter):
        self.base_filter = base_filter
        self.xor_filter = xor_filter
        self.merged_filter = (base_filter & ~xor_filter) | (~base_filter & xor_filter)

    def filter(self, update: Update) -> Optional[Union[bool, Dict]]:
        return self.merged_filter(update)

    @property
    def name(self) -> str:
        return f'<{self.base_filter} xor {self.xor_filter}>'

    @name.setter
    def name(self, name: str) -> NoReturn:
        raise RuntimeError('Cannot set name for XORFilter')


class _DiceEmoji(MessageFilter):
    def __init__(self, emoji: str = None, name: str = None):
        self.name = f'Filters.dice.{name}' if name else 'Filters.dice'
        self.emoji = emoji

    class _DiceValues(MessageFilter):
        def __init__(
            self,
            values: SLT[int],
            name: str,
            emoji: str = None,
        ):
            self.values = [values] if isinstance(values, int) else values
            self.emoji = emoji
            self.name = f'{name}({values})'

        def filter(self, message: Message) -> bool:
            if message.dice and message.dice.value in self.values:
                if self.emoji:
                    return message.dice.emoji == self.emoji
                return True
            return False

    def __call__(  # type: ignore[override]
        self, update: Union[Update, List[int], Tuple[int]]
    ) -> Union[bool, '_DiceValues']:
        if isinstance(update, Update):
            return self.filter(update.effective_message)
        return self._DiceValues(update, self.name, emoji=self.emoji)

    def filter(self, message: Message) -> bool:
        if bool(message.dice):
            if self.emoji:
                return message.dice.emoji == self.emoji
            return True
        return False


class Filters:
    """Predefined filters for use as the ``filter`` argument of
    :class:`telegram.ext.MessageHandler`.

    Examples:
        Use ``MessageHandler(Filters.video, callback_method)`` to filter all video
        messages. Use ``MessageHandler(Filters.contact, callback_method)`` for all contacts. etc.

    """

    class _All(MessageFilter):
        name = 'Filters.all'

        def filter(self, message: Message) -> bool:
            return True

    all = _All()
    """All Messages."""

    class _Text(MessageFilter):
        name = 'Filters.text'

        class _TextStrings(MessageFilter):
            def __init__(self, strings: Union[List[str], Tuple[str]]):
                self.strings = strings
                self.name = f'Filters.text({strings})'

            def filter(self, message: Message) -> bool:
                if message.text:
                    return message.text in self.strings
                return False

        def __call__(  # type: ignore[override]
            self, update: Union[Update, List[str], Tuple[str]]
        ) -> Union[bool, '_TextStrings']:
            if isinstance(update, Update):
                return self.filter(update.effective_message)
            return self._TextStrings(update)

        def filter(self, message: Message) -> bool:
            return bool(message.text)

    text = _Text()
    """Text Messages. If a list of strings is passed, it filters messages to only allow those
    whose text is appearing in the given list.

    Examples:
        To allow any text message, simply use
        ``MessageHandler(Filters.text, callback_method)``.

        A simple use case for passing a list is to allow only messages that were sent by a
        custom :class:`telegram.ReplyKeyboardMarkup`::

            buttons = ['Start', 'Settings', 'Back']
            markup = ReplyKeyboardMarkup.from_column(buttons)
            ...
            MessageHandler(Filters.text(buttons), callback_method)

    Note:
        * Dice messages don't have text. If you want to filter either text or dice messages, use
          ``Filters.text | Filters.dice``.
        * Messages containing a command are accepted by this filter. Use
          ``Filters.text & (~Filters.command)``, if you want to filter only text messages without
          commands.

    Args:
        update (List[:obj:`str`] | Tuple[:obj:`str`], optional): Which messages to allow. Only
            exact matches are allowed. If not specified, will allow any text message.
    """

    class _Caption(MessageFilter):
        name = 'Filters.caption'

        class _CaptionStrings(MessageFilter):
            def __init__(self, strings: Union[List[str], Tuple[str]]):
                self.strings = strings
                self.name = f'Filters.caption({strings})'

            def filter(self, message: Message) -> bool:
                if message.caption:
                    return message.caption in self.strings
                return False

        def __call__(  # type: ignore[override]
            self, update: Union[Update, List[str], Tuple[str]]
        ) -> Union[bool, '_CaptionStrings']:
            if isinstance(update, Update):
                return self.filter(update.effective_message)
            return self._CaptionStrings(update)

        def filter(self, message: Message) -> bool:
            return bool(message.caption)

    caption = _Caption()
    """Messages with a caption. If a list of strings is passed, it filters messages to only
    allow those whose caption is appearing in the given list.

    Examples:
        ``MessageHandler(Filters.caption, callback_method)``

    Args:
        update (List[:obj:`str`] | Tuple[:obj:`str`], optional): Which captions to allow. Only
            exact matches are allowed. If not specified, will allow any message with a caption.
    """

    class _Command(MessageFilter):
        name = 'Filters.command'

        class _CommandOnlyStart(MessageFilter):
            def __init__(self, only_start: bool):
                self.only_start = only_start
                self.name = f'Filters.command({only_start})'

            def filter(self, message: Message) -> bool:
                return bool(
                    message.entities
                    and any([e.type == MessageEntity.BOT_COMMAND for e in message.entities])
                )

        def __call__(  # type: ignore[override]
            self, update: Union[bool, Update]
        ) -> Union[bool, '_CommandOnlyStart']:
            if isinstance(update, Update):
                return self.filter(update.effective_message)
            return self._CommandOnlyStart(update)

        def filter(self, message: Message) -> bool:
            return bool(
                message.entities
                and message.entities[0].type == MessageEntity.BOT_COMMAND
                and message.entities[0].offset == 0
            )

    command = _Command()
    """
    Messages with a :attr:`telegram.MessageEntity.BOT_COMMAND`. By default only allows
    messages `starting` with a bot command. Pass :obj:`False` to also allow messages that contain a
    bot command `anywhere` in the text.

    Examples::

        MessageHandler(Filters.command, command_at_start_callback)
        MessageHandler(Filters.command(False), command_anywhere_callback)

    Note:
        ``Filters.text`` also accepts messages containing a command.

    Args:
        update (:obj:`bool`, optional): Whether to only allow messages that `start` with a bot
            command. Defaults to :obj:`True`.
    """

    class regex(MessageFilter):
        """
        Filters updates by searching for an occurrence of ``pattern`` in the message text.
        The ``re.search()`` function is used to determine whether an update should be filtered.

        Refer to the documentation of the ``re`` module for more information.

        To get the groups and groupdict matched, see :attr:`telegram.ext.CallbackContext.matches`.

        Examples:
            Use ``MessageHandler(Filters.regex(r'help'), callback)`` to capture all messages that
            contain the word 'help'. You can also use
            ``MessageHandler(Filters.regex(re.compile(r'help', re.IGNORECASE)), callback)`` if
            you want your pattern to be case insensitive. This approach is recommended
            if you need to specify flags on your pattern.

        Note:
            Filters use the same short circuiting logic as python's `and`, `or` and `not`.
            This means that for example:

                >>> Filters.regex(r'(a?x)') | Filters.regex(r'(b?x)')

            With a message.text of `x`, will only ever return the matches for the first filter,
            since the second one is never evaluated.

        Args:
            pattern (:obj:`str` | :obj:`Pattern`): The regex pattern.
        """

        data_filter = True

        def __init__(self, pattern: Union[str, Pattern]):
            if isinstance(pattern, str):
                pattern = re.compile(pattern)
            pattern = cast(Pattern, pattern)
            self.pattern: Pattern = pattern
            self.name = f'Filters.regex({self.pattern})'

        def filter(self, message: Message) -> Optional[Dict[str, List[Match]]]:
            """"""  # remove method from docs
            if message.text:
                match = self.pattern.search(message.text)
                if match:
                    return {'matches': [match]}
            return {}

    class caption_regex(MessageFilter):
        """
        Filters updates by searching for an occurrence of ``pattern`` in the message caption.

        This filter works similarly to :class:`Filters.regex`, with the only exception being that
        it applies to the message caption instead of the text.

        Examples:
            Use ``MessageHandler(Filters.photo & Filters.caption_regex(r'help'), callback)``
            to capture all photos with caption containing the word 'help'.

        Note:
            This filter will not work on simple text messages, but only on media with caption.

        Args:
            pattern (:obj:`str` | :obj:`Pattern`): The regex pattern.
        """

        data_filter = True

        def __init__(self, pattern: Union[str, Pattern]):
            if isinstance(pattern, str):
                pattern = re.compile(pattern)
            pattern = cast(Pattern, pattern)
            self.pattern: Pattern = pattern
            self.name = f'Filters.caption_regex({self.pattern})'

        def filter(self, message: Message) -> Optional[Dict[str, List[Match]]]:
            """"""  # remove method from docs
            if message.caption:
                match = self.pattern.search(message.caption)
                if match:
                    return {'matches': [match]}
            return {}

    class _Reply(MessageFilter):
        name = 'Filters.reply'

        def filter(self, message: Message) -> bool:
            return bool(message.reply_to_message)

    reply = _Reply()
    """Messages that are a reply to another message."""

    class _Audio(MessageFilter):
        name = 'Filters.audio'

        def filter(self, message: Message) -> bool:
            return bool(message.audio)

    audio = _Audio()
    """Messages that contain :class:`telegram.Audio`."""

    class _Document(MessageFilter):
        name = 'Filters.document'

        class category(MessageFilter):
            """Filters documents by their category in the mime-type attribute.

            Note:
                This Filter only filters by the mime_type of the document,
                    it doesn't check the validity of the document.
                The user can manipulate the mime-type of a message and
                    send media with wrong types that don't fit to this handler.

            Example:
                Filters.document.category('audio/') returns :obj:`True` for all types
                of audio sent as file, for example 'audio/mpeg' or 'audio/x-wav'.
            """

            def __init__(self, category: Optional[str]):
                """Initialize the category you want to filter

                Args:
                    category (str, optional): category of the media you want to filter"""
                self.category = category
                self.name = f"Filters.document.category('{self.category}')"

            def filter(self, message: Message) -> bool:
                """"""  # remove method from docs
                if message.document:
                    return message.document.mime_type.startswith(self.category)
                return False

        application = category('application/')
        audio = category('audio/')
        image = category('image/')
        video = category('video/')
        text = category('text/')

        class mime_type(MessageFilter):
            """This Filter filters documents by their mime-type attribute

            Note:
                This Filter only filters by the mime_type of the document,
                    it doesn't check the validity of document.
                The user can manipulate the mime-type of a message and
                    send media with wrong types that don't fit to this handler.

            Example:
                ``Filters.document.mime_type('audio/mpeg')`` filters all audio in mp3 format.
            """

            def __init__(self, mimetype: Optional[str]):
                """Initialize the category you want to filter

                Args:
                    mimetype (str, optional): mime_type of the media you want to filter"""
                self.mimetype = mimetype
                self.name = f"Filters.document.mime_type('{self.mimetype}')"

            def filter(self, message: Message) -> bool:
                """"""  # remove method from docs
                if message.document:
                    return message.document.mime_type == self.mimetype
                return False

        apk = mime_type('application/vnd.android.package-archive')
        doc = mime_type('application/msword')
        docx = mime_type('application/vnd.openxmlformats-officedocument.wordprocessingml.document')
        exe = mime_type('application/x-ms-dos-executable')
        gif = mime_type('video/mp4')
        jpg = mime_type('image/jpeg')
        mp3 = mime_type('audio/mpeg')
        pdf = mime_type('application/pdf')
        py = mime_type('text/x-python')
        svg = mime_type('image/svg+xml')
        txt = mime_type('text/plain')
        targz = mime_type('application/x-compressed-tar')
        wav = mime_type('audio/x-wav')
        xml = mime_type('application/xml')
        zip = mime_type('application/zip')

        class file_extension(MessageFilter):
            """This filter filters documents by their file ending/extension.

            Note:
                * This Filter only filters by the file ending/extension of the document,
                  it doesn't check the validity of document.
                * The user can manipulate the file extension of a document and
                  send media with wrong types that don't fit to this handler.
                * Case insensitive by default,
                  you may change this with the flag ``case_sensitive=True``.
                * Extension should be passed without leading dot
                  unless it's a part of the extension.
                * Pass :obj:`None` to filter files with no extension,
                  i.e. without a dot in the filename.

            Example:
                * ``Filters.document.file_extension("jpg")``
                  filters files with extension ``".jpg"``.
                * ``Filters.document.file_extension(".jpg")``
                  filters files with extension ``"..jpg"``.
                * ``Filters.document.file_extension("Dockerfile", case_sensitive=True)``
                  filters files with extension ``".Dockerfile"`` minding the case.
                * ``Filters.document.file_extension(None)``
                  filters files without a dot in the filename.
            """

            def __init__(self, file_extension: Optional[str], case_sensitive: bool = False):
                """Initialize the extension you want to filter.

                Args:
                    file_extension (:obj:`str` | :obj:`None`):
                        media file extension you want to filter.
                    case_sensitive (:obj:bool, optional):
                        pass :obj:`True` to make the filter case sensitive.
                        Default: :obj:`False`.
                """
                self.is_case_sensitive = case_sensitive
                if file_extension is None:
                    self.file_extension = None
                    self.name = "Filters.document.file_extension(None)"
                elif case_sensitive:
                    self.file_extension = f".{file_extension}"
                    self.name = (
                        f"Filters.document.file_extension({file_extension!r},"
                        " case_sensitive=True)"
                    )
                else:
                    self.file_extension = f".{file_extension}".lower()
                    self.name = f"Filters.document.file_extension({file_extension.lower()!r})"

            def filter(self, message: Message) -> bool:
                """"""  # remove method from docs
                if message.document is None:
                    return False
                if self.file_extension is None:
                    return "." not in message.document.file_name
                if self.is_case_sensitive:
                    filename = message.document.file_name
                else:
                    filename = message.document.file_name.lower()
                return filename.endswith(self.file_extension)

        def filter(self, message: Message) -> bool:
            return bool(message.document)

    document = _Document()
    """
    Subset for messages containing a document/file.

    Examples:
        Use these filters like: ``Filters.document.mp3``,
        ``Filters.document.mime_type("text/plain")`` etc. Or use just
        ``Filters.document`` for all document messages.

    Attributes:
        category: Filters documents by their category in the mime-type attribute

            Note:
                This Filter only filters by the mime_type of the document,
                it doesn't check the validity of the document.
                The user can manipulate the mime-type of a message and
                send media with wrong types that don't fit to this handler.

            Example:
                ``Filters.document.category('audio/')`` filters all types
                of audio sent as file, for example 'audio/mpeg' or 'audio/x-wav'.
        application: Same as ``Filters.document.category("application")``.
        audio: Same as ``Filters.document.category("audio")``.
        image: Same as ``Filters.document.category("image")``.
        video: Same as ``Filters.document.category("video")``.
        text: Same as ``Filters.document.category("text")``.
        mime_type: Filters documents by their mime-type attribute

            Note:
                This Filter only filters by the mime_type of the document,
                it doesn't check the validity of document.

                The user can manipulate the mime-type of a message and
                send media with wrong types that don't fit to this handler.

            Example:
                ``Filters.document.mime_type('audio/mpeg')`` filters all audio in mp3 format.
        apk: Same as ``Filters.document.mime_type("application/vnd.android.package-archive")``.
        doc: Same as ``Filters.document.mime_type("application/msword")``.
        docx: Same as ``Filters.document.mime_type("application/vnd.openxmlformats-\
officedocument.wordprocessingml.document")``.
        exe: Same as ``Filters.document.mime_type("application/x-ms-dos-executable")``.
        gif: Same as ``Filters.document.mime_type("video/mp4")``.
        jpg: Same as ``Filters.document.mime_type("image/jpeg")``.
        mp3: Same as ``Filters.document.mime_type("audio/mpeg")``.
        pdf: Same as ``Filters.document.mime_type("application/pdf")``.
        py: Same as ``Filters.document.mime_type("text/x-python")``.
        svg: Same as ``Filters.document.mime_type("image/svg+xml")``.
        txt: Same as ``Filters.document.mime_type("text/plain")``.
        targz: Same as ``Filters.document.mime_type("application/x-compressed-tar")``.
        wav: Same as ``Filters.document.mime_type("audio/x-wav")``.
        xml: Same as ``Filters.document.mime_type("application/xml")``.
        zip: Same as ``Filters.document.mime_type("application/zip")``.
        file_extension: This filter filters documents by their file ending/extension.

            Note:
                * This Filter only filters by the file ending/extension of the document,
                  it doesn't check the validity of document.
                * The user can manipulate the file extension of a document and
                  send media with wrong types that don't fit to this handler.
                * Case insensitive by default,
                  you may change this with the flag ``case_sensitive=True``.
                * Extension should be passed without leading dot
                  unless it's a part of the extension.
                * Pass :obj:`None` to filter files with no extension,
                  i.e. without a dot in the filename.

            Example:
                * ``Filters.document.file_extension("jpg")``
                  filters files with extension ``".jpg"``.
                * ``Filters.document.file_extension(".jpg")``
                  filters files with extension ``"..jpg"``.
                * ``Filters.document.file_extension("Dockerfile", case_sensitive=True)``
                  filters files with extension ``".Dockerfile"`` minding the case.
                * ``Filters.document.file_extension(None)``
                  filters files without a dot in the filename.
    """

    class _Animation(MessageFilter):
        name = 'Filters.animation'

        def filter(self, message: Message) -> bool:
            return bool(message.animation)

    animation = _Animation()
    """Messages that contain :class:`telegram.Animation`."""

    class _Photo(MessageFilter):
        name = 'Filters.photo'

        def filter(self, message: Message) -> bool:
            return bool(message.photo)

    photo = _Photo()
    """Messages that contain :class:`telegram.PhotoSize`."""

    class _Sticker(MessageFilter):
        name = 'Filters.sticker'

        def filter(self, message: Message) -> bool:
            return bool(message.sticker)

    sticker = _Sticker()
    """Messages that contain :class:`telegram.Sticker`."""

    class _Video(MessageFilter):
        name = 'Filters.video'

        def filter(self, message: Message) -> bool:
            return bool(message.video)

    video = _Video()
    """Messages that contain :class:`telegram.Video`."""

    class _Voice(MessageFilter):
        name = 'Filters.voice'

        def filter(self, message: Message) -> bool:
            return bool(message.voice)

    voice = _Voice()
    """Messages that contain :class:`telegram.Voice`."""

    class _VideoNote(MessageFilter):
        name = 'Filters.video_note'

        def filter(self, message: Message) -> bool:
            return bool(message.video_note)

    video_note = _VideoNote()
    """Messages that contain :class:`telegram.VideoNote`."""

    class _Contact(MessageFilter):
        name = 'Filters.contact'

        def filter(self, message: Message) -> bool:
            return bool(message.contact)

    contact = _Contact()
    """Messages that contain :class:`telegram.Contact`."""

    class _Location(MessageFilter):
        name = 'Filters.location'

        def filter(self, message: Message) -> bool:
            return bool(message.location)

    location = _Location()
    """Messages that contain :class:`telegram.Location`."""

    class _Venue(MessageFilter):
        name = 'Filters.venue'

        def filter(self, message: Message) -> bool:
            return bool(message.venue)

    venue = _Venue()
    """Messages that contain :class:`telegram.Venue`."""

    class _StatusUpdate(UpdateFilter):
        """Subset for messages containing a status update.

        Examples:
            Use these filters like: ``Filters.status_update.new_chat_members`` etc. Or use just
            ``Filters.status_update`` for all status update messages.

        """

        class _NewChatMembers(MessageFilter):
            name = 'Filters.status_update.new_chat_members'

            def filter(self, message: Message) -> bool:
                return bool(message.new_chat_members)

        new_chat_members = _NewChatMembers()
        """Messages that contain :attr:`telegram.Message.new_chat_members`."""

        class _LeftChatMember(MessageFilter):
            name = 'Filters.status_update.left_chat_member'

            def filter(self, message: Message) -> bool:
                return bool(message.left_chat_member)

        left_chat_member = _LeftChatMember()
        """Messages that contain :attr:`telegram.Message.left_chat_member`."""

        class _NewChatTitle(MessageFilter):
            name = 'Filters.status_update.new_chat_title'

            def filter(self, message: Message) -> bool:
                return bool(message.new_chat_title)

        new_chat_title = _NewChatTitle()
        """Messages that contain :attr:`telegram.Message.new_chat_title`."""

        class _NewChatPhoto(MessageFilter):
            name = 'Filters.status_update.new_chat_photo'

            def filter(self, message: Message) -> bool:
                return bool(message.new_chat_photo)

        new_chat_photo = _NewChatPhoto()
        """Messages that contain :attr:`telegram.Message.new_chat_photo`."""

        class _DeleteChatPhoto(MessageFilter):
            name = 'Filters.status_update.delete_chat_photo'

            def filter(self, message: Message) -> bool:
                return bool(message.delete_chat_photo)

        delete_chat_photo = _DeleteChatPhoto()
        """Messages that contain :attr:`telegram.Message.delete_chat_photo`."""

        class _ChatCreated(MessageFilter):
            name = 'Filters.status_update.chat_created'

            def filter(self, message: Message) -> bool:
                return bool(
                    message.group_chat_created
                    or message.supergroup_chat_created
                    or message.channel_chat_created
                )

        chat_created = _ChatCreated()
        """Messages that contain :attr:`telegram.Message.group_chat_created`,
            :attr: `telegram.Message.supergroup_chat_created` or
            :attr: `telegram.Message.channel_chat_created`."""

        class _Migrate(MessageFilter):
            name = 'Filters.status_update.migrate'

            def filter(self, message: Message) -> bool:
                return bool(message.migrate_from_chat_id or message.migrate_to_chat_id)

        migrate = _Migrate()
        """Messages that contain :attr:`telegram.Message.migrate_from_chat_id` or
            :attr:`telegram.Message.migrate_to_chat_id`."""

        class _PinnedMessage(MessageFilter):
            name = 'Filters.status_update.pinned_message'

            def filter(self, message: Message) -> bool:
                return bool(message.pinned_message)

        pinned_message = _PinnedMessage()
        """Messages that contain :attr:`telegram.Message.pinned_message`."""

        class _ConnectedWebsite(MessageFilter):
            name = 'Filters.status_update.connected_website'

            def filter(self, message: Message) -> bool:
                return bool(message.connected_website)

        connected_website = _ConnectedWebsite()
        """Messages that contain :attr:`telegram.Message.connected_website`."""

        class _ProximityAlertTriggered(MessageFilter):
            name = 'Filters.status_update.proximity_alert_triggered'

            def filter(self, message: Message) -> bool:
                return bool(message.proximity_alert_triggered)

        proximity_alert_triggered = _ProximityAlertTriggered()
        """Messages that contain :attr:`telegram.Message.proximity_alert_triggered`."""

        name = 'Filters.status_update'

        def filter(self, message: Update) -> bool:
            return bool(
                self.new_chat_members(message)
                or self.left_chat_member(message)
                or self.new_chat_title(message)
                or self.new_chat_photo(message)
                or self.delete_chat_photo(message)
                or self.chat_created(message)
                or self.migrate(message)
                or self.pinned_message(message)
                or self.connected_website(message)
                or self.proximity_alert_triggered(message)
            )

    status_update = _StatusUpdate()
    """Subset for messages containing a status update.

    Examples:
        Use these filters like: ``Filters.status_update.new_chat_members`` etc. Or use just
        ``Filters.status_update`` for all status update messages.

    Attributes:
        chat_created: Messages that contain
            :attr:`telegram.Message.group_chat_created`,
            :attr:`telegram.Message.supergroup_chat_created` or
            :attr:`telegram.Message.channel_chat_created`.
        connected_website: Messages that contain
            :attr:`telegram.Message.connected_website`.
        delete_chat_photo: Messages that contain
            :attr:`telegram.Message.delete_chat_photo`.
        left_chat_member: Messages that contain
            :attr:`telegram.Message.left_chat_member`.
        migrate: Messages that contain
            :attr:`telegram.Message.migrate_from_chat_id` or
            :attr: `telegram.Message.migrate_from_chat_id`.
        new_chat_members: Messages that contain
            :attr:`telegram.Message.new_chat_members`.
        new_chat_photo: Messages that contain
            :attr:`telegram.Message.new_chat_photo`.
        new_chat_title: Messages that contain
            :attr:`telegram.Message.new_chat_title`.
        pinned_message: Messages that contain
            :attr:`telegram.Message.pinned_message`.
        proximity_alert_triggered: Messages that contain
            :attr:`telegram.Message.proximity_alert_triggered`.
    """

    class _Forwarded(MessageFilter):
        name = 'Filters.forwarded'

        def filter(self, message: Message) -> bool:
            return bool(message.forward_date)

    forwarded = _Forwarded()
    """Messages that are forwarded."""

    class _Game(MessageFilter):
        name = 'Filters.game'

        def filter(self, message: Message) -> bool:
            return bool(message.game)

    game = _Game()
    """Messages that contain :class:`telegram.Game`."""

    class entity(MessageFilter):
        """
        Filters messages to only allow those which have a :class:`telegram.MessageEntity`
        where their `type` matches `entity_type`.

        Examples:
            Example ``MessageHandler(Filters.entity("hashtag"), callback_method)``

        Args:
            entity_type: Entity type to check for. All types can be found as constants
                in :class:`telegram.MessageEntity`.

        """

        def __init__(self, entity_type: str):
            self.entity_type = entity_type
            self.name = f'Filters.entity({self.entity_type})'

        def filter(self, message: Message) -> bool:
            """"""  # remove method from docs
            return any(entity.type == self.entity_type for entity in message.entities)

    class caption_entity(MessageFilter):
        """
        Filters media messages to only allow those which have a :class:`telegram.MessageEntity`
        where their `type` matches `entity_type`.

        Examples:
            Example ``MessageHandler(Filters.caption_entity("hashtag"), callback_method)``

        Args:
            entity_type: Caption Entity type to check for. All types can be found as constants
                in :class:`telegram.MessageEntity`.

        """

        def __init__(self, entity_type: str):
            self.entity_type = entity_type
            self.name = f'Filters.caption_entity({self.entity_type})'

        def filter(self, message: Message) -> bool:
            """"""  # remove method from docs
            return any(entity.type == self.entity_type for entity in message.caption_entities)

    class _Private(MessageFilter):
        name = 'Filters.private'

        def filter(self, message: Message) -> bool:
            warnings.warn(
                'Filters.private is deprecated. Use Filters.chat_type.private instead.',
                TelegramDeprecationWarning,
                stacklevel=2,
            )
            return message.chat.type == Chat.PRIVATE

    private = _Private()
    """
    Messages sent in a private chat.

    Note:
        DEPRECATED. Use
        :attr:`telegram.ext.Filters.chat_type.private` instead.
    """

    class _Group(MessageFilter):
        name = 'Filters.group'

        def filter(self, message: Message) -> bool:
            warnings.warn(
                'Filters.group is deprecated. Use Filters.chat_type.groups instead.',
                TelegramDeprecationWarning,
                stacklevel=2,
            )
            return message.chat.type in [Chat.GROUP, Chat.SUPERGROUP]

    group = _Group()
    """
    Messages sent in a group or a supergroup chat.

    Note:
        DEPRECATED. Use
        :attr:`telegram.ext.Filters.chat_type.groups` instead.
    """

    class _ChatType(MessageFilter):
        name = 'Filters.chat_type'

        class _Channel(MessageFilter):
            name = 'Filters.chat_type.channel'

            def filter(self, message: Message) -> bool:
                return message.chat.type == Chat.CHANNEL

        channel = _Channel()

        class _Group(MessageFilter):
            name = 'Filters.chat_type.group'

            def filter(self, message: Message) -> bool:
                return message.chat.type == Chat.GROUP

        group = _Group()

        class _SuperGroup(MessageFilter):
            name = 'Filters.chat_type.supergroup'

            def filter(self, message: Message) -> bool:
                return message.chat.type == Chat.SUPERGROUP

        supergroup = _SuperGroup()

        class _Groups(MessageFilter):
            name = 'Filters.chat_type.groups'

            def filter(self, message: Message) -> bool:
                return message.chat.type in [Chat.GROUP, Chat.SUPERGROUP]

        groups = _Groups()

        class _Private(MessageFilter):
            name = 'Filters.chat_type.private'

            def filter(self, message: Message) -> bool:
                return message.chat.type == Chat.PRIVATE

        private = _Private()

        def filter(self, message: Message) -> bool:
            return bool(message.chat.type)

    chat_type = _ChatType()
    """Subset for filtering the type of chat.

    Examples:
        Use these filters like: ``Filters.chat_type.channel`` or
        ``Filters.chat_type.supergroup`` etc. Or use just ``Filters.chat_type`` for all
        chat types.

    Attributes:
        channel: Updates from channel
        group: Updates from group
        supergroup: Updates from supergroup
        groups: Updates from group *or* supergroup
        private: Updates sent in private chat
    """

    class _ChatUserBaseFilter(MessageFilter):
        def __init__(
            self,
            chat_id: SLT[int] = None,
            username: SLT[str] = None,
            allow_empty: bool = False,
        ):
            self.chat_id_name = 'chat_id'
            self.username_name = 'username'
            self.allow_empty = allow_empty
            self.__lock = Lock()

            self._chat_ids: Set[int] = set()
            self._usernames: Set[str] = set()

            self._set_chat_ids(chat_id)
            self._set_usernames(username)

        @abstractmethod
        def get_chat_or_user(self, message: Message) -> Union[Chat, User, None]:
            pass

        @staticmethod
        def _parse_chat_id(chat_id: SLT[int]) -> Set[int]:
            if chat_id is None:
                return set()
            if isinstance(chat_id, int):
                return {chat_id}
            return set(chat_id)

        @staticmethod
        def _parse_username(username: SLT[str]) -> Set[str]:
            if username is None:
                return set()
            if isinstance(username, str):
                return {username[1:] if username.startswith('@') else username}
            return {chat[1:] if chat.startswith('@') else chat for chat in username}

        def _set_chat_ids(self, chat_id: SLT[int]) -> None:
            with self.__lock:
                if chat_id and self._usernames:
                    raise RuntimeError(
                        f"Can't set {self.chat_id_name} in conjunction with (already set) "
                        f"{self.username_name}s."
                    )
                self._chat_ids = self._parse_chat_id(chat_id)

        def _set_usernames(self, username: SLT[str]) -> None:
            with self.__lock:
                if username and self._chat_ids:
                    raise RuntimeError(
                        f"Can't set {self.username_name} in conjunction with (already set) "
                        f"{self.chat_id_name}s."
                    )
                self._usernames = self._parse_username(username)

        @property
        def chat_ids(self) -> FrozenSet[int]:
            with self.__lock:
                return frozenset(self._chat_ids)

        @chat_ids.setter
        def chat_ids(self, chat_id: SLT[int]) -> None:
            self._set_chat_ids(chat_id)

        @property
        def usernames(self) -> FrozenSet[str]:
            with self.__lock:
                return frozenset(self._usernames)

        @usernames.setter
        def usernames(self, username: SLT[str]) -> None:
            self._set_usernames(username)

        def add_usernames(self, username: SLT[str]) -> None:
            with self.__lock:
                if self._chat_ids:
                    raise RuntimeError(
                        f"Can't set {self.username_name} in conjunction with (already set) "
                        f"{self.chat_id_name}s."
                    )

                parsed_username = self._parse_username(username)
                self._usernames |= parsed_username

        def add_chat_ids(self, chat_id: SLT[int]) -> None:
            with self.__lock:
                if self._usernames:
                    raise RuntimeError(
                        f"Can't set {self.chat_id_name} in conjunction with (already set) "
                        f"{self.username_name}s."
                    )

                parsed_chat_id = self._parse_chat_id(chat_id)

                self._chat_ids |= parsed_chat_id

        def remove_usernames(self, username: SLT[str]) -> None:
            with self.__lock:
                if self._chat_ids:
                    raise RuntimeError(
                        f"Can't set {self.username_name} in conjunction with (already set) "
                        f"{self.chat_id_name}s."
                    )

                parsed_username = self._parse_username(username)
                self._usernames -= parsed_username

        def remove_chat_ids(self, chat_id: SLT[int]) -> None:
            with self.__lock:
                if self._usernames:
                    raise RuntimeError(
                        f"Can't set {self.chat_id_name} in conjunction with (already set) "
                        f"{self.username_name}s."
                    )
                parsed_chat_id = self._parse_chat_id(chat_id)
                self._chat_ids -= parsed_chat_id

        def filter(self, message: Message) -> bool:
            """"""  # remove method from docs
            chat_or_user = self.get_chat_or_user(message)
            if chat_or_user:
                if self.chat_ids:
                    return chat_or_user.id in self.chat_ids
                if self.usernames:
                    return bool(chat_or_user.username and chat_or_user.username in self.usernames)
                return self.allow_empty
            return False

        @property
        def name(self) -> str:
            return (
                f'Filters.{self.__class__.__name__}('
                f'{", ".join(str(s) for s in (self.usernames or self.chat_ids))})'
            )

        @name.setter
        def name(self, name: str) -> NoReturn:
            raise RuntimeError(f'Cannot set name for Filters.{self.__class__.__name__}')

    class user(_ChatUserBaseFilter):
        # pylint: disable=W0235
        """Filters messages to allow only those which are from specified user ID(s) or
        username(s).

        Examples:
            ``MessageHandler(Filters.user(1234), callback_method)``

        Warning:
            :attr:`user_ids` will give a *copy* of the saved user ids as :class:`frozenset`. This
            is to ensure thread safety. To add/remove a user, you should use :meth:`add_usernames`,
            :meth:`add_user_ids`, :meth:`remove_usernames` and :meth:`remove_user_ids`. Only update
            the entire set by ``filter.user_ids/usernames = new_set``, if you are entirely sure
            that it is not causing race conditions, as this will complete replace the current set
            of allowed users.

        Args:
            user_id(:class:`telegram.utils.types.SLT[int]`, optional):
                Which user ID(s) to allow through.
            username(:class:`telegram.utils.types.SLT[str]`, optional):
                Which username(s) to allow through. Leading ``'@'`` s in usernames will be
                discarded.
            allow_empty(:obj:`bool`, optional): Whether updates should be processed, if no user
                is specified in :attr:`user_ids` and :attr:`usernames`. Defaults to :obj:`False`

        Raises:
            RuntimeError: If user_id and username are both present.

        Attributes:
            user_ids(set(:obj:`int`), optional): Which user ID(s) to allow through.
            usernames(set(:obj:`str`), optional): Which username(s) (without leading ``'@'``) to
                allow through.
            allow_empty(:obj:`bool`, optional): Whether updates should be processed, if no user
                is specified in :attr:`user_ids` and :attr:`usernames`.

        """

        def __init__(
            self,
            user_id: SLT[int] = None,
            username: SLT[str] = None,
            allow_empty: bool = False,
        ):
            super().__init__(chat_id=user_id, username=username, allow_empty=allow_empty)
            self.chat_id_name = 'user_id'

        def get_chat_or_user(self, message: Message) -> Optional[User]:
            return message.from_user

        @property
        def user_ids(self) -> FrozenSet[int]:
            return self.chat_ids

        @user_ids.setter
        def user_ids(self, user_id: SLT[int]) -> None:
            self.chat_ids = user_id  # type: ignore[assignment]

        def add_usernames(self, username: SLT[str]) -> None:
            """
            Add one or more users to the allowed usernames.

            Args:
                username(:class:`telegram.utils.types.SLT[str]`, optional):
                    Which username(s) to allow through.
                    Leading ``'@'`` s in usernames will be discarded.
            """
            return super().add_usernames(username)

        def add_user_ids(self, user_id: SLT[int]) -> None:
            """
            Add one or more users to the allowed user ids.

            Args:
                user_id(:class:`telegram.utils.types.SLT[int]`, optional):
                    Which user ID(s) to allow through.
            """
            return super().add_chat_ids(user_id)

        def remove_usernames(self, username: SLT[str]) -> None:
            """
            Remove one or more users from allowed usernames.

            Args:
                username(:class:`telegram.utils.types.SLT[str]`, optional):
                    Which username(s) to disallow through.
                    Leading ``'@'`` s in usernames will be discarded.
            """
            return super().remove_usernames(username)

        def remove_user_ids(self, user_id: SLT[int]) -> None:
            """
            Remove one or more users from allowed user ids.

            Args:
                user_id(:class:`telegram.utils.types.SLT[int]`, optional):
                    Which user ID(s) to disallow through.
            """
            return super().remove_chat_ids(user_id)

    class via_bot(_ChatUserBaseFilter):
        # pylint: disable=W0235
        """Filters messages to allow only those which are from specified via_bot ID(s) or
        username(s).

        Examples:
            ``MessageHandler(Filters.via_bot(1234), callback_method)``

        Warning:
            :attr:`bot_ids` will give a *copy* of the saved bot ids as :class:`frozenset`. This
            is to ensure thread safety. To add/remove a bot, you should use :meth:`add_usernames`,
            :meth:`add_bot_ids`, :meth:`remove_usernames` and :meth:`remove_bot_ids`. Only update
            the entire set by ``filter.bot_ids/usernames = new_set``, if you are entirely sure
            that it is not causing race conditions, as this will complete replace the current set
            of allowed bots.

        Args:
            bot_id(:class:`telegram.utils.types.SLT[int]`, optional):
                Which bot ID(s) to allow through.
            username(:class:`telegram.utils.types.SLT[str]`, optional):
                Which username(s) to allow through. Leading ``'@'`` s in usernames will be
                discarded.
            allow_empty(:obj:`bool`, optional): Whether updates should be processed, if no user
                is specified in :attr:`bot_ids` and :attr:`usernames`. Defaults to :obj:`False`

        Raises:
            RuntimeError: If bot_id and username are both present.

        Attributes:
            bot_ids(set(:obj:`int`), optional): Which bot ID(s) to allow through.
            usernames(set(:obj:`str`), optional): Which username(s) (without leading ``'@'``) to
                allow through.
            allow_empty(:obj:`bool`, optional): Whether updates should be processed, if no bot
                is specified in :attr:`bot_ids` and :attr:`usernames`.

        """

        def __init__(
            self,
            bot_id: SLT[int] = None,
            username: SLT[str] = None,
            allow_empty: bool = False,
        ):
            super().__init__(chat_id=bot_id, username=username, allow_empty=allow_empty)
            self.chat_id_name = 'bot_id'

        def get_chat_or_user(self, message: Message) -> Optional[User]:
            return message.via_bot

        @property
        def bot_ids(self) -> FrozenSet[int]:
            return self.chat_ids

        @bot_ids.setter
        def bot_ids(self, bot_id: SLT[int]) -> None:
            self.chat_ids = bot_id  # type: ignore[assignment]

        def add_usernames(self, username: SLT[str]) -> None:
            """
            Add one or more users to the allowed usernames.

            Args:
                username(:class:`telegram.utils.types.SLT[str]`, optional):
                    Which username(s) to allow through.
                    Leading ``'@'`` s in usernames will be discarded.
            """
            return super().add_usernames(username)

        def add_bot_ids(self, bot_id: SLT[int]) -> None:
            """

            Add one or more users to the allowed user ids.

            Args:
                bot_id(:class:`telegram.utils.types.SLT[int]`, optional):
                    Which bot ID(s) to allow through.
            """
            return super().add_chat_ids(bot_id)

        def remove_usernames(self, username: SLT[str]) -> None:
            """
            Remove one or more users from allowed usernames.

            Args:
                username(:class:`telegram.utils.types.SLT[str]`, optional):
                    Which username(s) to disallow through.
                    Leading ``'@'`` s in usernames will be discarded.
            """
            return super().remove_usernames(username)

        def remove_bot_ids(self, bot_id: SLT[int]) -> None:
            """
            Remove one or more users from allowed user ids.

            Args:
                bot_id(:class:`telegram.utils.types.SLT[int]`, optional):
                    Which bot ID(s) to disallow through.
            """
            return super().remove_chat_ids(bot_id)

    class chat(_ChatUserBaseFilter):
        # pylint: disable=W0235
        """Filters messages to allow only those which are from a specified chat ID or username.

        Examples:
            ``MessageHandler(Filters.chat(-1234), callback_method)``

        Warning:
            :attr:`chat_ids` will give a *copy* of the saved chat ids as :class:`frozenset`. This
            is to ensure thread safety. To add/remove a chat, you should use :meth:`add_usernames`,
            :meth:`add_chat_ids`, :meth:`remove_usernames` and :meth:`remove_chat_ids`. Only update
            the entire set by ``filter.chat_ids/usernames = new_set``, if you are entirely sure
            that it is not causing race conditions, as this will complete replace the current set
            of allowed chats.

        Args:
            chat_id(:class:`telegram.utils.types.SLT[int]`, optional):
                Which chat ID(s) to allow through.
            username(:class:`telegram.utils.types.SLT[str]`, optional):
                Which username(s) to allow through.
                Leading ``'@'`` s in usernames will be discarded.
            allow_empty(:obj:`bool`, optional): Whether updates should be processed, if no chat
                is specified in :attr:`chat_ids` and :attr:`usernames`. Defaults to :obj:`False`

        Raises:
            RuntimeError: If chat_id and username are both present.

        Attributes:
            chat_ids(set(:obj:`int`), optional): Which chat ID(s) to allow through.
            usernames(set(:obj:`str`), optional): Which username(s) (without leading ``'@'``) to
                allow through.
            allow_empty(:obj:`bool`, optional): Whether updates should be processed, if no chat
                is specified in :attr:`chat_ids` and :attr:`usernames`.

        """

        def get_chat_or_user(self, message: Message) -> Optional[Chat]:
            return message.chat

        def add_usernames(self, username: SLT[str]) -> None:
            """
            Add one or more chats to the allowed usernames.

            Args:
                username(:class:`telegram.utils.types.SLT[str]`, optional):
                    Which username(s) to allow through.
                    Leading ``'@'`` s in usernames will be discarded.
            """
            return super().add_usernames(username)

        def add_chat_ids(self, chat_id: SLT[int]) -> None:
            """
            Add one or more chats to the allowed chat ids.

            Args:
                chat_id(:class:`telegram.utils.types.SLT[int]`, optional):
                    Which chat ID(s) to allow through.
            """
            return super().add_chat_ids(chat_id)

        def remove_usernames(self, username: SLT[str]) -> None:
            """
            Remove one or more chats from allowed usernames.

            Args:
                username(:class:`telegram.utils.types.SLT[str]`, optional):
                    Which username(s) to disallow through.
                    Leading ``'@'`` s in usernames will be discarded.
            """
            return super().remove_usernames(username)

        def remove_chat_ids(self, chat_id: SLT[int]) -> None:
            """
            Remove one or more chats from allowed chat ids.

            Args:
                chat_id(:class:`telegram.utils.types.SLT[int]`, optional):
                    Which chat ID(s) to disallow through.
            """
            return super().remove_chat_ids(chat_id)

    class sender_chat(_ChatUserBaseFilter):
        # pylint: disable=W0235
        """Filters messages to allow only those which are from a specified sender chats chat ID or
        username.

        Examples:
            * To filter for messages forwarded from a channel with ID ``-1234``, use
              ``MessageHandler(Filters.sender_chat(-1234), callback_method)``.
            * To filter for messages of anonymous admins in a super group with username
              ``@anonymous``, use
              ``MessageHandler(Filters.sender_chat(username='anonymous'), callback_method)``.
            * To filter for messages forwarded from *any* channel, use
              ``MessageHandler(Filters.sender_chat.channel, callback_method)``.
            * To filter for messages of anonymous admins in *any* super group, use
              ``MessageHandler(Filters.sender_chat.super_group, callback_method)``.

        Warning:
            :attr:`chat_ids` will return a *copy* of the saved chat ids as :class:`frozenset`. This
            is to ensure thread safety. To add/remove a chat, you should use :meth:`add_usernames`,
            :meth:`add_chat_ids`, :meth:`remove_usernames` and :meth:`remove_chat_ids`. Only update
            the entire set by ``filter.chat_ids/usernames = new_set``, if you are entirely sure
            that it is not causing race conditions, as this will complete replace the current set
            of allowed chats.

        Args:
            chat_id(:class:`telegram.utils.types.SLT[int]`, optional):
                Which sender chat chat ID(s) to allow through.
            username(:class:`telegram.utils.types.SLT[str]`, optional):
                Which sender chat username(s) to allow through.
                Leading ``'@'`` s in usernames will be discarded.
            allow_empty(:obj:`bool`, optional): Whether updates should be processed, if no sender
                chat is specified in :attr:`chat_ids` and :attr:`usernames`. Defaults to
                :obj:`False`

        Raises:
            RuntimeError: If both chat_id and username are present.

        Attributes:
            chat_ids(set(:obj:`int`), optional): Which sender chat chat ID(s) to allow through.
            usernames(set(:obj:`str`), optional): Which sender chat username(s) (without leading
                ``'@'``) to allow through.
            allow_empty(:obj:`bool`, optional): Whether updates should be processed, if no sender
                chat is specified in :attr:`chat_ids` and :attr:`usernames`.
            super_group: Messages whose sender chat is a super group.

                Examples:
                    ``Filters.sender_chat.supergroup``
            channel: Messages whose sender chat is a channel.

                Examples:
                    ``Filters.sender_chat.channel``

        """

        def get_chat_or_user(self, message: Message) -> Optional[Chat]:
            return message.sender_chat

        def add_usernames(self, username: SLT[str]) -> None:
            """
            Add one or more sender chats to the allowed usernames.

            Args:
                username(:class:`telegram.utils.types.SLT[str]`, optional):
                    Which sender chat username(s) to allow through.
                    Leading ``'@'`` s in usernames will be discarded.
            """
            return super().add_usernames(username)

        def add_chat_ids(self, chat_id: SLT[int]) -> None:
            """
            Add one or more sender chats to the allowed chat ids.

            Args:
                chat_id(:class:`telegram.utils.types.SLT[int]`, optional):
                    Which sender chat ID(s) to allow through.
            """
            return super().add_chat_ids(chat_id)

        def remove_usernames(self, username: SLT[str]) -> None:
            """
            Remove one or more sender chats from allowed usernames.

            Args:
                username(:class:`telegram.utils.types.SLT[str]`, optional):
                    Which sender chat username(s) to disallow through.
                    Leading ``'@'`` s in usernames will be discarded.
            """
            return super().remove_usernames(username)

        def remove_chat_ids(self, chat_id: SLT[int]) -> None:
            """
            Remove one or more sender chats from allowed chat ids.

            Args:
                chat_id(:class:`telegram.utils.types.SLT[int]`, optional):
                    Which sender chat ID(s) to disallow through.
            """
            return super().remove_chat_ids(chat_id)

        class _SuperGroup(MessageFilter):
            def filter(self, message: Message) -> bool:
                if message.sender_chat:
                    return message.sender_chat.type == Chat.SUPERGROUP
                return False

        class _Channel(MessageFilter):
            def filter(self, message: Message) -> bool:
                if message.sender_chat:
                    return message.sender_chat.type == Chat.CHANNEL
                return False

        super_group = _SuperGroup()
        channel = _Channel()

    class _Invoice(MessageFilter):
        name = 'Filters.invoice'

        def filter(self, message: Message) -> bool:
            return bool(message.invoice)

    invoice = _Invoice()
    """Messages that contain :class:`telegram.Invoice`."""

    class _SuccessfulPayment(MessageFilter):
        name = 'Filters.successful_payment'

        def filter(self, message: Message) -> bool:
            return bool(message.successful_payment)

    successful_payment = _SuccessfulPayment()
    """Messages that confirm a :class:`telegram.SuccessfulPayment`."""

    class _PassportData(MessageFilter):
        name = 'Filters.passport_data'

        def filter(self, message: Message) -> bool:
            return bool(message.passport_data)

    passport_data = _PassportData()
    """Messages that contain a :class:`telegram.PassportData`"""

    class _Poll(MessageFilter):
        name = 'Filters.poll'

        def filter(self, message: Message) -> bool:
            return bool(message.poll)

    poll = _Poll()
    """Messages that contain a :class:`telegram.Poll`."""

    class _Dice(_DiceEmoji):
        dice = _DiceEmoji('🎲', 'dice')
        darts = _DiceEmoji('🎯', 'darts')
        basketball = _DiceEmoji('🏀', 'basketball')
        football = _DiceEmoji('⚽')
        slot_machine = _DiceEmoji('🎰')

    dice = _Dice()
    """Dice Messages. If an integer or a list of integers is passed, it filters messages to only
    allow those whose dice value is appearing in the given list.

    Examples:
        To allow any dice message, simply use
        ``MessageHandler(Filters.dice, callback_method)``.
        To allow only dice with value 6, use
        ``MessageHandler(Filters.dice(6), callback_method)``.
        To allow only dice with value 5 `or` 6, use
        ``MessageHandler(Filters.dice([5, 6]), callback_method)``.

    Note:
        Dice messages don't have text. If you want to filter either text or dice messages, use
        ``Filters.text | Filters.dice``.

    Args:
        update (:class:`telegram.utils.types.SLT[int]`, optional):
            Which values to allow. If not specified, will allow any dice message.

    Attributes:
        dice: Dice messages with the emoji 🎲. Passing a list of integers is supported just as for
            :attr:`Filters.dice`.
        darts: Dice messages with the emoji 🎯. Passing a list of integers is supported just as for
            :attr:`Filters.dice`.
        basketball: Dice messages with the emoji 🏀. Passing a list of integers is supported just
            as for :attr:`Filters.dice`.
        football: Dice messages with the emoji ⚽. Passing a list of integers is supported just
            as for :attr:`Filters.dice`.
        slot_machine: Dice messages with the emoji 🎰. Passing a list of integers is supported just
            as for :attr:`Filters.dice`.
    """

    class language(MessageFilter):
        """Filters messages to only allow those which are from users with a certain language code.

        Note:
            According to official Telegram API documentation, not every single user has the
            `language_code` attribute. Do not count on this filter working on all users.

        Examples:
            ``MessageHandler(Filters.language("en"), callback_method)``

        Args:
            lang (:class:`telegram.utils.types.SLT[str]`):
                Which language code(s) to allow through.
                This will be matched using ``.startswith`` meaning that
                'en' will match both 'en_US' and 'en_GB'.

        """

        def __init__(self, lang: SLT[str]):
            if isinstance(lang, str):
                lang = cast(str, lang)
                self.lang = [lang]
            else:
                lang = cast(List[str], lang)
                self.lang = lang
            self.name = f'Filters.language({self.lang})'

        def filter(self, message: Message) -> bool:
            """"""  # remove method from docs
            return bool(
                message.from_user.language_code
                and any([message.from_user.language_code.startswith(x) for x in self.lang])
            )

    class _UpdateType(UpdateFilter):
        name = 'Filters.update'

        class _Message(UpdateFilter):
            name = 'Filters.update.message'

            def filter(self, update: Update) -> bool:
                return update.message is not None

        message = _Message()

        class _EditedMessage(UpdateFilter):
            name = 'Filters.update.edited_message'

            def filter(self, update: Update) -> bool:
                return update.edited_message is not None

        edited_message = _EditedMessage()

        class _Messages(UpdateFilter):
            name = 'Filters.update.messages'

            def filter(self, update: Update) -> bool:
                return update.message is not None or update.edited_message is not None

        messages = _Messages()

        class _ChannelPost(UpdateFilter):
            name = 'Filters.update.channel_post'

            def filter(self, update: Update) -> bool:
                return update.channel_post is not None

        channel_post = _ChannelPost()

        class _EditedChannelPost(UpdateFilter):
            name = 'Filters.update.edited_channel_post'

            def filter(self, update: Update) -> bool:
                return update.edited_channel_post is not None

        edited_channel_post = _EditedChannelPost()

        class _ChannelPosts(UpdateFilter):
            name = 'Filters.update.channel_posts'

            def filter(self, update: Update) -> bool:
                return update.channel_post is not None or update.edited_channel_post is not None

        channel_posts = _ChannelPosts()

        def filter(self, update: Update) -> bool:
            return bool(self.messages(update) or self.channel_posts(update))

    update = _UpdateType()
    """Subset for filtering the type of update.

    Examples:
        Use these filters like: ``Filters.update.message`` or
        ``Filters.update.channel_posts`` etc. Or use just ``Filters.update`` for all
        types.

    Attributes:
        message: Updates with :attr:`telegram.Update.message`
        edited_message: Updates with :attr:`telegram.Update.edited_message`
        messages: Updates with either :attr:`telegram.Update.message` or
            :attr:`telegram.Update.edited_message`
        channel_post: Updates with :attr:`telegram.Update.channel_post`
        edited_channel_post: Updates with
            :attr:`telegram.Update.edited_channel_post`
        channel_posts: Updates with either :attr:`telegram.Update.channel_post` or
            :attr:`telegram.Update.edited_channel_post`
    """



#!/usr/bin/env python
#
# A library that provides a Python interface to the Telegram Bot API
# Copyright (C) 2015-2021
# Leandro Toledo de Souza <devs@python-telegram-bot.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].

import pytest

from telegram import EncryptedCredentials, PassportElementError


@pytest.fixture(scope='class')
def encrypted_credentials():
    return EncryptedCredentials(
        TestEncryptedCredentials.data,
        TestEncryptedCredentials.hash,
        TestEncryptedCredentials.secret,
    )


class TestEncryptedCredentials:
    data = 'data'
    hash = 'hash'
    secret = 'secret'

    def test_expected_values(self, encrypted_credentials):
        assert encrypted_credentials.data == self.data
        assert encrypted_credentials.hash == self.hash
        assert encrypted_credentials.secret == self.secret

    def test_to_dict(self, encrypted_credentials):
        encrypted_credentials_dict = encrypted_credentials.to_dict()

        assert isinstance(encrypted_credentials_dict, dict)
        assert encrypted_credentials_dict['data'] == encrypted_credentials.data
        assert encrypted_credentials_dict['hash'] == encrypted_credentials.hash
        assert encrypted_credentials_dict['secret'] == encrypted_credentials.secret

    def test_equality(self):
        a = EncryptedCredentials(self.data, self.hash, self.secret)
        b = EncryptedCredentials(self.data, self.hash, self.secret)
        c = EncryptedCredentials(self.data, '', '')
        d = EncryptedCredentials('', self.hash, '')
        e = EncryptedCredentials('', '', self.secret)
        f = PassportElementError('source', 'type', 'message')

        assert a == b
        assert hash(a) == hash(b)
        assert a is not b

        assert a != c
        assert hash(a) != hash(c)

        assert a != d
        assert hash(a) != hash(d)

        assert a != e
        assert hash(a) != hash(e)

        assert a != f
        assert hash(a) != hash(f)



#!/usr/bin/env python
#
# A library that provides a Python interface to the Telegram Bot API
# Copyright (C) 2015-2021
# Leandro Toledo de Souza <devs@python-telegram-bot.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].

import pytest

from telegram import GameHighScore, User


@pytest.fixture(scope='class')
def game_highscore():
    return GameHighScore(
        TestGameHighScore.position, TestGameHighScore.user, TestGameHighScore.score
    )


class TestGameHighScore:
    position = 12
    user = User(2, 'test user', False)
    score = 42

    def test_de_json(self, bot):
        json_dict = {'position': self.position, 'user': self.user.to_dict(), 'score': self.score}
        highscore = GameHighScore.de_json(json_dict, bot)

        assert highscore.position == self.position
        assert highscore.user == self.user
        assert highscore.score == self.score

    def test_to_dict(self, game_highscore):
        game_highscore_dict = game_highscore.to_dict()

        assert isinstance(game_highscore_dict, dict)
        assert game_highscore_dict['position'] == game_highscore.position
        assert game_highscore_dict['user'] == game_highscore.user.to_dict()
        assert game_highscore_dict['score'] == game_highscore.score

    def test_equality(self):
        a = GameHighScore(1, User(2, 'test user', False), 42)
        b = GameHighScore(1, User(2, 'test user', False), 42)
        c = GameHighScore(2, User(2, 'test user', False), 42)
        d = GameHighScore(1, User(3, 'test user', False), 42)
        e = User(3, 'test user', False)

        assert a == b
        assert hash(a) == hash(b)

        assert a != c
        assert hash(a) != hash(c)

        assert a != d
        assert hash(a) != hash(d)

        assert a != e
        assert hash(a) != hash(e)
